/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdint.h>

#include "network_state.h"

NetworkState::NetworkState()
{
	Create();
}

NetworkState::~NetworkState()
{
	Destroy();
}

error_e NetworkState::AddRoute(const char *ip, const char *mask, const char *interface,
const char *gateway, int family, gboolean *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		ADD_ROUTE,
		g_variant_new("(ssssi)", ip, mask, interface,
			gateway, family),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e NetworkState::RemoveRoute(const char *ip, const char *mask, const char *interface,
const char *gateway, int family, gboolean *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		REMOVE_ROUTE,
		g_variant_new("(ssssi)", ip, mask, interface,
			gateway, family),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::EthernetCableState(int *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		ETHERNET_CABLE_STATE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e NetworkState::LaunchMdns(const char *name)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		LAUNCH_MDNS,
		g_variant_new("(s)", name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::DevicePolicySetWifi(int state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		DEVICE_POLICY_SET_WIFI,
		g_variant_new("(i)", state),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::DevicePolicyGetWifi(int *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		DEVICE_POLICY_GET_WIFI,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::DevicePolicySetWifiProfile(int state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		DEVICE_POLICY_SET_WIFI_PROFILE,
		g_variant_new("(i)", state),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::DevicePolicyGetWifiProfile(int *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		DEVICE_POLICY_GET_WIFI_PROFILE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::GetMeteredInfo(gboolean *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		GET_METERED_INFO,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::PreferredIpv6Address(const char *profile, char *address)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_NETWORK_INTERFACE,
		PREFERRED_IPV6_ADDRESS,
		g_variant_new("(s)", profile),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", address);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::GetTCPDumpState(gboolean *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_TCPDUMP_INTERFACE,
		GET_TCP_DUMPSTATE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::StartTCPDump()
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_TCPDUMP_INTERFACE,
		START_TCP_DUMP,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
error_e NetworkState::StopTCPDump()
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_NETWORK_PATH,
		NETCONFIG_TCPDUMP_INTERFACE,
		STOP_TCP_DUMP,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
