/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <gmock/gmock.h>
#include <gtest/gtest.h>
#include <stdint.h>
#include <glib.h>

#include "wifi.h"

Wifi::Wifi()
{
	Create();
}

Wifi::~Wifi()
{
	Destroy();
}

error_e Wifi::GetWifiState(char *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_WIFI_STATE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::IsIpConflictDetectEnabled(gboolean *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		IS_IP_CONFLICT_DETECT_ENABLED,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SetBgscan(int scan_mode)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SET_BGSCAN,
		g_variant_new("(u)", scan_mode),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetAutoscan(gboolean *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_AUTOSCAN,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetIpConflictState(int *state)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_IP_CONFLICT_STATE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(u)", state);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetIpConflictPeriod(int *period)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_IP_CONFLICT_PERIOD,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(u)", period);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetAutoscanmode(int *autoscanmode)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_AUTOSCANMODE,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(u)", autoscanmode);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::NetlinkScan(const char *ssid)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GVariantBuilder *builder;
	GVariant *params = NULL;

	if (ssid[0] == '\0')
		return ERROR_INVALID_PARAMETER;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(builder, "{sv}", "SSID", g_variant_new_string(ssid));

	params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		NETLINK_SCAN,
		params,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::RequestWpsConnect(const char *param)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REQUEST_WPS_CONNECT,
		g_variant_new("(s)", param),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
error_e Wifi::GetPasspoint(int *enable)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_PASSPOINT,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", enable);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SetPasspoint(int enable)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SET_PASSPOINT,
		g_variant_new("(i)", enable),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetConfigIds(void)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_CONFIGIDS,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SaveConfiguration(const char *config_id, const char *name,
		const char *ssid, const char *passphrase)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	GVariantBuilder *b = NULL;
	GVariant *params = NULL;

	b = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(b, "{sv}", "Name", g_variant_new_string(name));
	g_variant_builder_add(b, "{sv}", "SSID", g_variant_new_string(ssid));
	if (passphrase != NULL)
		g_variant_builder_add(b, "{sv}", "Passphrase", g_variant_new_string(passphrase));
	params = g_variant_new("(s@a{sv})", config_id, g_variant_builder_end(b));
	g_variant_builder_unref(b);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SAVE_CONFIGURATION,
		params,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SaveEapConfiguration(const char *config_id, const char *name,
		const char *ssid)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	GVariantBuilder *b = NULL;
	GVariant *params = NULL;

	b = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	g_variant_builder_add(b, "{sv}", "Name", g_variant_new_string(name));
	g_variant_builder_add(b, "{sv}", "SSID", g_variant_new_string(ssid));
	params = g_variant_new("(s@a{sv})", config_id, g_variant_builder_end(b));
	g_variant_builder_unref(b);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SAVE_EAP_CONFIGURATION,
		params,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SetIpConflictPeriod(unsigned int time)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SET_IP_CONFLICT_PERIOD,
		g_variant_new("(u)", time),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::RemoveConfiguration(const char *config_id)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REMOVE_CONFIGURATION,
		g_variant_new("(s)", config_id),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::LoadConfiguration(const char *config_id)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		LOAD_CONFIGURATION,
		g_variant_new("(s)", config_id),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::LoadEapConfiguration(const char *config_id)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		LOAD_EAP_CONFIGURATION,
		g_variant_new("(s)", config_id),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::SetConfigField(const char *config_id, const char *key,
		const char *value)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		SET_CONFIG_FIELD,
		g_variant_new("(sss)", config_id, key, value),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetConfigPassphrase(const char *config_id, char *password)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_CONFIG_PASSPHRASE,
		g_variant_new("(s)", config_id),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", password);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::CreateEapConfig(const char *profile_name)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GVariant *params = NULL;
	GVariantBuilder *builder;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{ss}"));
	g_variant_builder_add(builder, "{ss}", "Type", "wifi");
	g_variant_builder_add(builder, "{ss}",
			"Name", "gtest");

	params = g_variant_new("(o@a{ss})", (gchar *)profile_name,
			g_variant_builder_end(builder));

	g_variant_builder_unref(builder);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		CREATE_EAP_CONFIG,
		params,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::DeleteEapConfig(const char *profile_name)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		DELETE_EAP_CONFIG,
		g_variant_new("(s)", profile_name),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::IpConflictSetEnable(bool detect)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		IP_CONFLICT_SET_ENABLE,
		g_variant_new("(b)", detect),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetSimImsi(char *imsi)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_SIM_IMSI,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", imsi);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::ReqSimAuth(const unsigned char *rand_data, unsigned int length,
	gboolean *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GVariantBuilder *b;
	GVariant *v;
	b = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for(unsigned int i = 0;i < length;i++)
		g_variant_builder_add(b, "y", rand_data[i]);

	v = g_variant_new("(ay)", b);
	g_variant_builder_unref(b);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REQ_SIM_AUTH,
		v,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::ReqAkaAuth(const unsigned char *rand_data, int rand_length,
	const unsigned char *autn_data, int auth_length, gboolean *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GVariantBuilder *b1, *b2;
	GVariant *v;
	b1 = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for(int i = 0;i < rand_length;i++)
		g_variant_builder_add(b1, "y", rand_data[i]);

	b2 = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for(int i = 0;i < auth_length;i++)
		g_variant_builder_add(b2, "y", autn_data[i]);

	v = g_variant_new("(@ay@ay)", g_variant_builder_end(b1), g_variant_builder_end(b2));

	g_variant_builder_unref(b1);
	g_variant_builder_unref(b2);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REQ_AKA_AUTH,
		v,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetSimAuth(const unsigned char *auth_data, unsigned int length)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;
	GVariantBuilder *b;
	GVariant *v;
	b = g_variant_builder_new(G_VARIANT_TYPE("ay"));
	for(unsigned int i = 0;i < length;i++)
		g_variant_builder_add(b, "y", auth_data[i]);

	v = g_variant_new("(ay)", b);
	g_variant_builder_unref(b);

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_SIM_AUTH,
		v,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::CheckBlackList(const char *name, const char *sec_type,
		const char *eap, gboolean *allowed)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		CHECK_BLACKLIST,
		g_variant_new("(sss)", name, sec_type, eap),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(b)", allowed);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsDisconnect(const char *peer_mac_address, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_DISCONNECT,
		g_variant_new("(s)", peer_mac_address),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsConnectedPeer(char *peer_mac_address)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_CONNECTED_PEER,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", peer_mac_address);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsConnect(const char *peer_mac_address, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_CONNECT,
		g_variant_new("(s)", peer_mac_address),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsDiscover(const char *peer_mac_address, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_DISCOVER,
		g_variant_new("(s)", peer_mac_address),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsChannelSwitch(const char *peer_mac_address, int freq, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_CHANNEL_SWITCH,
		g_variant_new("(si)", peer_mac_address, freq),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::TdlsCancelChannelSwitch(const char *peer_mac_address, int *result)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		TDLS_CANCEL_CHANNEL_SWITCH,
		g_variant_new("(s)", peer_mac_address),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(i)", result);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::AddVsie(int frame_id, const char *vsie)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		ADD_VSIE,
		g_variant_new("(is)", frame_id, vsie),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::GetVsie(int frame_id, char *vsie)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		GET_VSIE,
		g_variant_new("(i)", frame_id),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", vsie);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::RemoveVsie(int frame_id, const char *vsie)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REMOVE_VSIE,
		g_variant_new("(is)", frame_id, vsie),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::EncryptPassphrase(char **enc_data, const char *passphrase)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		ENCRYPT_PASSPHRASE,
		g_variant_new("(s)", passphrase),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", enc_data);
	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::DecryptPassphrase(const char *enc_data, char **passphrase)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		DECRYPT_PASSPHRASE,
		g_variant_new("(s)", enc_data),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_get(message, "(s)", passphrase);
	g_variant_unref(message);

	return ERROR_NONE;
}


error_e Wifi::Stop(const char *device)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_FIRMWARE_INTERFACE,
		WIFI_FIRMWARE_STOP,
		g_variant_new("(s)", device),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::Start(const char *device)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_FIRMWARE_INTERFACE,
		WIFI_FIRMWARE_START,
		g_variant_new("(s)", device),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::LoadDriver(bool device_picker_test)
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		LOAD_DRIVER,
		g_variant_new("(b)", device_picker_test),
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}

error_e Wifi::RemoveDriver()
{
	GVariant *message = NULL;
	error_e error = ERROR_NONE;

	message = InvokeMethod(NETCONFIG_SERVICE,
		NETCONFIG_WIFI_PATH,
		NETCONFIG_WIFI_INTERFACE,
		REMOVE_DRIVER,
		NULL,
		&error);

	if (message == NULL) {
		GLOGD("Failed to invoke dbus method");
		return error;
	}

	g_variant_unref(message);

	return ERROR_NONE;
}
