/*
 * Network Configuration Module
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "plugin.h"
#include "util.h"

#ifdef USE_NETCONFIG_LOG
#include "log.h"
#else
#include <dlog.h>

#define NETCONFIG_TAG		"NETCONFIG"
#define __LOG(level, format, arg...) \
	do { \
		SLOG(level, NETCONFIG_TAG, format, ## arg); \
	} while (0)

#define DBG(format, arg...)	__LOG(LOG_DEBUG, format, ## arg)
#define ERR(format, arg...)	__LOG(LOG_ERROR, format, ## arg)
#endif

#define BATTERY_MODE_END   0
#define BATTERY_MODE_GET   1

typedef struct {
	time_t time;
} battery_scan_data_s;

typedef struct {
	time_t time;
	int rssi;
} battery_rssi_data_s;

static battery_scan_data_s g_wifi_scan_data = { 0, };

static battery_rssi_data_s g_dn_rssi_data = { 0, };
static battery_rssi_data_s g_wifi_rssi_data = { 0, };

static battery_dn_data_s *g_dn_data = NULL;
static battery_wifi_data_s *g_wifi_data = NULL;

static GSList *g_dn_list = NULL;
static GSList *g_wifi_list = NULL;

static void __battery_free_app_data(void *data)
{
	battery_app_data_s *app_data = (battery_app_data_s *)data;

	g_free(app_data->app_id);
	g_free(app_data);
}

static void __battery_free_dn_data(void *data)
{
	battery_dn_data_s *dn_data = (battery_dn_data_s *)data;

	g_slist_free_full(dn_data->atm_list, __battery_free_app_data);
	g_free(dn_data);
}

static void __battery_flush_dn_list(void)
{
	g_slist_free_full(g_dn_list, __battery_free_dn_data);
	g_dn_list = NULL;
}

static void __battery_free_wifi_data(void *data)
{
	battery_wifi_data_s *wifi_data = (battery_wifi_data_s *)data;

	g_slist_free_full(wifi_data->atm_list, __battery_free_app_data);
	g_free(wifi_data);
}

static void __battery_flush_wifi_list(void)
{
	g_slist_free_full(g_wifi_list, __battery_free_wifi_data);
	g_wifi_list = NULL;
}

static void __battery_print_dn_list(void)
{
	GSList *list = NULL;

	for (list = g_dn_list; list != NULL; list = list->next) {
		battery_dn_data_s *dn_data = (battery_dn_data_s *)list->data;
		DBG("[Battery DN] rssi[%d/%d/%d/%d/%d/%d/%d]",
			dn_data->time_level_0, dn_data->time_level_1, dn_data->time_level_2,
			dn_data->time_level_3, dn_data->time_level_4, dn_data->time_level_5,
			dn_data->time_level_6);
	}
}

static void __battery_print_wifi_list(void)
{
	GSList* list = NULL;

	for (list = g_wifi_list; list != NULL; list = list->next) {
		battery_wifi_data_s *wifi_data = (battery_wifi_data_s *)list->data;
		DBG("[Battery Wi-Fi] start[%ld] end[%ld] scan[%d] rssi[%d/%d/%d/%d/%d]",
			wifi_data->start_time, wifi_data->end_time, wifi_data->scan_time,
			wifi_data->time_level_0, wifi_data->time_level_1, wifi_data->time_level_2,
			wifi_data->time_level_3, wifi_data->time_level_4);
	}
}

static void __battery_get_dn_list(void *data, battery_dn_data_s *dn_data)
{
	GVariantBuilder *builder = (GVariantBuilder *)data;
	GVariantBuilder sub_builder;

	g_variant_builder_init(&sub_builder, G_VARIANT_TYPE("a{sv}"));

	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_0,
		g_variant_new_uint32(dn_data->time_level_0));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_1,
		g_variant_new_uint32(dn_data->time_level_1));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_2,
		g_variant_new_uint32(dn_data->time_level_2));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_3,
		g_variant_new_uint32(dn_data->time_level_3));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_4,
		g_variant_new_uint32(dn_data->time_level_4));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_5,
		g_variant_new_uint32(dn_data->time_level_5));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_DN_TIME_LEVEL_6,
		g_variant_new_uint32(dn_data->time_level_6));

	DBG("[Battery DN] rssi[%d/%d/%d/%d/%d/%d/%d]",
			dn_data->time_level_0, dn_data->time_level_1, dn_data->time_level_2,
			dn_data->time_level_3, dn_data->time_level_4, dn_data->time_level_5,
			dn_data->time_level_6);

	if (dn_data->atm_list != NULL) {
		GSList *list = NULL;
		GVariantBuilder atm_builder;

		g_variant_builder_init(&atm_builder, G_VARIANT_TYPE("aa{sv}"));

		for (list = dn_data->atm_list; list != NULL; list = list->next) {
			battery_app_data_s *app_data = (battery_app_data_s *)list->data;

			g_variant_builder_open(&atm_builder, G_VARIANT_TYPE("a{sv}"));

			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_APP_ID,
				g_variant_new_string(app_data->app_id));
			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_RX,
				g_variant_new_uint32(app_data->rx));
			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_TX,
				g_variant_new_uint32(app_data->tx));

			g_variant_builder_close(&atm_builder);

			DBG("[Battery DN] app_id[%s] rx[%d] tx[%d]", app_data->app_id, app_data->rx, app_data->tx);
		}

		g_variant_builder_add(&sub_builder, "{sv}", BATTERY_ATM_LIST,
						g_variant_new_variant(g_variant_builder_end(&atm_builder)));
	}

	g_variant_builder_add_value(builder, g_variant_builder_end(&sub_builder));
}

static void __battery_get_wifi_list(void *data, battery_wifi_data_s *wifi_data)
{
	GVariantBuilder *builder = (GVariantBuilder *)data;
	GVariantBuilder sub_builder;

	g_variant_builder_init(&sub_builder, G_VARIANT_TYPE("a{sv}"));

	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_TIME_LEVEL_0,
		g_variant_new_uint32(wifi_data->time_level_0));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_TIME_LEVEL_1,
		g_variant_new_uint32(wifi_data->time_level_1));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_TIME_LEVEL_2,
		g_variant_new_uint32(wifi_data->time_level_2));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_TIME_LEVEL_3,
		g_variant_new_uint32(wifi_data->time_level_3));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_TIME_LEVEL_4,
		g_variant_new_uint32(wifi_data->time_level_4));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_START_TIME,
		g_variant_new_uint32(wifi_data->start_time));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_END_TIME,
		g_variant_new_uint32(wifi_data->end_time));
	g_variant_builder_add(&sub_builder, "{sv}", BATTERY_WIFI_SCAN_TIME,
		g_variant_new_uint32(wifi_data->scan_time));

	DBG("[Battery Wi-Fi] start[%ld] end[%ld] scan[%d] rssi[%d/%d/%d/%d/%d]",
			wifi_data->start_time, wifi_data->end_time, wifi_data->scan_time,
			wifi_data->time_level_0, wifi_data->time_level_1, wifi_data->time_level_2,
			wifi_data->time_level_3, wifi_data->time_level_4);


	if (wifi_data->atm_list != NULL) {
		GSList *list = NULL;
		GVariantBuilder atm_builder;

		g_variant_builder_init(&atm_builder, G_VARIANT_TYPE("aa{sv}"));

		for (list = wifi_data->atm_list; list != NULL; list = list->next) {
			battery_app_data_s *app_data = (battery_app_data_s *)list->data;

			g_variant_builder_open(&atm_builder, G_VARIANT_TYPE("a{sv}"));

			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_APP_ID,
				g_variant_new_string(app_data->app_id));
			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_RX,
				g_variant_new_uint32(app_data->rx));
			g_variant_builder_add(&atm_builder, "{sv}", BATTERY_ATM_TX,
				g_variant_new_uint32(app_data->tx));

			g_variant_builder_close(&atm_builder);

			DBG("[Battery Wi-Fi] app_id[%s] rx[%d] tx[%d]", app_data->app_id, app_data->rx, app_data->tx);
		}

		g_variant_builder_add(&sub_builder, "{sv}", BATTERY_ATM_LIST,
						g_variant_new_variant(g_variant_builder_end(&atm_builder)));
	}

	g_variant_builder_add_value(builder, g_variant_builder_end(&sub_builder));
}

void battery_update_dn_rssi(int rssi)
{
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_dn_data == NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	DBG("[Battery DN] rssi(%ld) [%d]", curr_time, rssi);

	if (g_dn_rssi_data.time != 0) {
		uint diff_time = curr_time - g_dn_rssi_data.time;
		diff_time = (diff_time > 0) ? diff_time : 1;
		switch (g_dn_rssi_data.rssi) {
		case 0:
			g_dn_data->time_level_0 += diff_time;
			DBG("[Battery DN] level 0 time [%d/%d]", diff_time, g_dn_data->time_level_0);
			break;
		case 1:
			g_dn_data->time_level_1 += diff_time;
			DBG("[Battery DN] level 1 time [%d/%d]", diff_time, g_dn_data->time_level_1);
			break;
		case 2:
			g_dn_data->time_level_2 += diff_time;
			DBG("[Battery DN] level 2 time [%d/%d]", diff_time, g_dn_data->time_level_2);
			break;
		case 3:
			g_dn_data->time_level_3 += diff_time;
			DBG("[Battery DN] level 3 time [%d/%d]", diff_time, g_dn_data->time_level_3);
			break;
		case 4:
			g_dn_data->time_level_4 += diff_time;
			DBG("[Battery DN] level 4 time [%d/%d]", diff_time, g_dn_data->time_level_4);
			break;
		case 5:
			g_dn_data->time_level_5 += diff_time;
			DBG("[Battery DN] level 5 time [%d/%d]", diff_time, g_dn_data->time_level_5);
			break;
		case 6:
			g_dn_data->time_level_6 += diff_time;
			DBG("[Battery DN] level 6 time [%d/%d]", diff_time, g_dn_data->time_level_6);
			break;
		default:
			break;
		}
	}

	if (rssi >= 0) {
		g_dn_rssi_data.time = curr_time;
		g_dn_rssi_data.rssi = rssi;
	}
}

void battery_start_dn_data(void)
{
	battery_dn_data_s *dn_data = NULL;
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_dn_data != NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	__battery_print_dn_list();

	dn_data = (battery_dn_data_s *)g_try_malloc0(sizeof(battery_dn_data_s));
	if (dn_data == NULL)
		return;

	memset(dn_data, 0, sizeof(battery_dn_data_s));
	dn_data->start_time = curr_time;

	g_dn_data = dn_data;
	g_dn_list = g_slist_append(g_dn_list, g_dn_data);

	if (g_dn_rssi_data.time != 0)
		g_dn_rssi_data.time = curr_time;

	DBG("[Battery DN] start(%ld)", g_dn_data->start_time);
}

void battery_end_dn_data(int mode, stc_get_stats_cb get_stats_cb)
{
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_dn_data == NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	battery_update_dn_rssi(-1);

	if (mode == BATTERY_MODE_END) {
		g_dn_rssi_data.time = 0;
		g_dn_rssi_data.rssi = 0;
	}

	g_dn_data->end_time = curr_time;

	get_stats_cb(g_dn_data->start_time, g_dn_data->end_time, &(g_dn_data->atm_list));

	g_dn_data = NULL;

	DBG("[Battery DN] end");
}

void battery_update_wifi_scan(int state)
{
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_wifi_data == NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	switch (state) {
	case 0: /* scan done */
		DBG("[Battery Wi-Fi] scan(%ld) [done]", curr_time);
		if (g_wifi_scan_data.time != 0) {
			uint diff_time = curr_time - g_wifi_scan_data.time;
			diff_time = (diff_time > 0) ? diff_time : 1;
			g_wifi_data->scan_time += diff_time;
			DBG("[Battery Wi-Fi] scan time [%d/%d]", diff_time, g_wifi_data->scan_time);
		}
		break;
	case 1: /* scan started */
		g_wifi_scan_data.time = curr_time;
		DBG("[Battery Wi-Fi] scan(%ld) [started]", curr_time);
		break;
	default:
		break;
	}
}

void battery_update_wifi_rssi(int rssi)
{
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_wifi_data == NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	DBG("[Battery Wi-Fi] rssi(%ld) [%d]", curr_time, rssi);

	if (g_wifi_rssi_data.time != 0) {
		uint diff_time = curr_time - g_wifi_rssi_data.time;
		diff_time = (diff_time > 0) ? diff_time : 1;
		switch (g_wifi_rssi_data.rssi) {
		case 0:
			g_wifi_data->time_level_0 += diff_time;
			DBG("[Battery Wi-Fi] level 0 time [%d/%d]", diff_time, g_wifi_data->time_level_0);
			break;
		case 1:
			g_wifi_data->time_level_1 += diff_time;
			DBG("[Battery Wi-Fi] level 1 time [%d/%d]", diff_time, g_wifi_data->time_level_1);
			break;
		case 2:
			g_wifi_data->time_level_2 += diff_time;
			DBG("[Battery Wi-Fi] level 2 time [%d/%d]", diff_time, g_wifi_data->time_level_2);
			break;
		case 3:
			g_wifi_data->time_level_3 += diff_time;
			DBG("[Battery Wi-Fi] level 3 time [%d/%d]", diff_time, g_wifi_data->time_level_3);
			break;
		case 4:
			g_wifi_data->time_level_4 += diff_time;
			DBG("[Battery Wi-Fi] level 4 time [%d/%d]", diff_time, g_wifi_data->time_level_4);
			break;
		default:
			break;
		}
	}

	if (rssi >= 0) {
		g_wifi_rssi_data.time = curr_time;
		g_wifi_rssi_data.rssi = rssi;
	}
}

void battery_start_wifi_data(void)
{
	battery_wifi_data_s *wifi_data = NULL;
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_wifi_data != NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	__battery_print_wifi_list();

	wifi_data = (battery_wifi_data_s *)g_try_malloc0(sizeof(battery_wifi_data_s));
	if (wifi_data == NULL)
		return;

	memset(wifi_data, 0, sizeof(battery_wifi_data_s));
	wifi_data->start_time = curr_time;

	g_wifi_data = wifi_data;
	g_wifi_list = g_slist_append(g_wifi_list, g_wifi_data);

	if (g_wifi_rssi_data.time)
		g_wifi_rssi_data.time = curr_time;

	DBG("[Battery Wi-Fi] start(%ld)", g_wifi_data->start_time);
}

void battery_end_wifi_data(int mode, stc_get_stats_cb get_stats_cb)
{
	struct timespec curr_timespec;
	time_t curr_time;

	if (g_wifi_data == NULL)
		return;

	clock_gettime(CLOCK_MONOTONIC, &curr_timespec);
	curr_time = curr_timespec.tv_sec;

	battery_update_wifi_rssi(-1);

	if (mode == BATTERY_MODE_END) {
		g_wifi_rssi_data.time = 0;
		g_wifi_rssi_data.rssi = 0;
	}

	g_wifi_scan_data.time = 0;

	g_wifi_data->end_time = curr_time;

	get_stats_cb(g_wifi_data->start_time, g_wifi_data->end_time, &(g_wifi_data->atm_list));

	g_wifi_data = NULL;

	DBG("[Battery Wi-Fi] end(%ld)", curr_time);
}

void battery_get_dn_list(void *data, stc_get_stats_cb get_stats_cb)
{
	GSList *dn_list = NULL;
	int cont = FALSE;

	if (g_dn_data != NULL)
		cont = TRUE;

	battery_end_dn_data(BATTERY_MODE_GET, get_stats_cb);

	DBG("===============================================================");

	for (dn_list = g_dn_list; dn_list != NULL; dn_list = dn_list->next) {
		battery_dn_data_s *dn_data = (battery_dn_data_s *)dn_list->data;
		__battery_get_dn_list(data, dn_data);
		DBG("===============================================================");
	}

	__battery_flush_dn_list();

	if (cont)
		battery_start_dn_data();
}

void battery_get_wifi_list(void *data, stc_get_stats_cb get_stats_cb)
{
	GSList *list = NULL;
	int cont = FALSE;

	if (g_wifi_data != NULL)
		cont = TRUE;

	battery_end_wifi_data(BATTERY_MODE_GET, get_stats_cb);

	DBG("===============================================================");

	for (list = g_wifi_list; list != NULL; list = list->next) {
		battery_wifi_data_s *wifi_data = (battery_wifi_data_s *)list->data;
		__battery_get_wifi_list(data, wifi_data);
		DBG("===============================================================");
	}

	__battery_flush_wifi_list();

	if (cont)
		battery_start_wifi_data();
}

extern struct netconfig_battery_plugin_t netconfig_battery_plugin
		__attribute__ ((visibility("default")));
struct netconfig_battery_plugin_t netconfig_battery_plugin = {
	battery_start_dn_data,
	battery_end_dn_data,
	battery_update_dn_rssi,
	battery_start_wifi_data,
	battery_end_wifi_data,
	battery_update_wifi_scan,
	battery_update_wifi_rssi,
	battery_get_dn_list,
	battery_get_wifi_list
};
