/*
 * Network Configuration Module
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <app.h>
#include <errno.h>
#include <vconf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <vconf-keys.h>
#include <syspopup_caller.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <eventsystem.h>
#include <tzplatform_config.h>
#include <system_info.h>

#include "plugin.h"

#ifdef USE_NETCONFIG_LOG
#include "log.h"
#else
#include <dlog.h>

#define NETCONFIG_TAG		"NETCONFIG"
#define __LOG(level, format, arg...) \
	do { \
		SLOG(level, NETCONFIG_TAG, format, ## arg); \
	} while (0)

#define DBG(format, arg...)	__LOG(LOG_DEBUG, format, ## arg)
#define ERR(format, arg...)	__LOG(LOG_ERROR, format, ## arg)
#endif

const char *_headed_convert_sys_event(int event_val)
{
	switch (event_val) {
	case SYS_EVT_NETWORK_STATUS:
		return SYS_EVENT_NETWORK_STATUS;
	case SYS_EVT_WIFI_STATE:
			return SYS_EVENT_WIFI_STATE;
	case EKEY_NETWORK_STATUS:
			return EVT_KEY_NETWORK_STATUS;
	case EKEY_WIFI_STATE:
			return EVT_KEY_WIFI_STATE;
	case EVAL_NETWORK_WIFI:
			return EVT_VAL_NETWORK_WIFI;
	case EVAL_NETWORK_CELLULAR:
			return EVT_VAL_NETWORK_CELLULAR;
	case EVAL_NETWORK_ETHERNET:
			return EVT_VAL_NETWORK_ETHERNET;
	case EVAL_NETWORK_BT:
			return EVT_VAL_NETWORK_BT;
	case EVAL_NETWORK_DISCONNECTED:
			return EVT_VAL_NETWORK_DISCONNECTED;
	case EVAL_WIFI_CONNECTED:
			return EVT_VAL_WIFI_CONNECTED;
	case EVAL_WIFI_ON:
			return EVT_VAL_WIFI_ON;
	case EVAL_WIFI_OFF:
			return EVT_VAL_WIFI_OFF;
	default:
		break;
	}

	return NULL;
}

void headed_pop_device_picker(void)
{
#if defined TIZEN_WEARABLE
	int ret = 0;
	app_control_h	control = NULL;

	ret = app_control_create(&control);
	if (APP_CONTROL_ERROR_NONE != ret) {
		DBG("failed to create app control");
		return ;
	}

	app_control_add_extra_data(control, "viewtype", "scanlist");

	app_control_set_app_id(control, "org.tizen.wifi");
	ret = app_control_send_launch_request(control, NULL, NULL);
	if (APP_CONTROL_ERROR_NONE == ret)
		DBG("Launch request sent successfully");

	app_control_destroy(control);
#else
	bundle *b = NULL;
	int wifi_ug_state = 0;
	int ret = 0;

	ret = vconf_get_int(VCONFKEY_WIFI_UG_RUN_STATE, &wifi_ug_state);
	if (ret != VCONF_OK)
		ERR("Failed to get vconfkey [%s] value", VCONFKEY_WIFI_UG_RUN_STATE);

	if (wifi_ug_state == VCONFKEY_WIFI_UG_RUN_STATE_ON_FOREGROUND)
		return;

	b = bundle_create();

	DBG("Launch Wi-Fi device picker");
	syspopup_launch("wifi-qs", b);

	bundle_free(b);
#endif
}


gboolean headed_send_notification_to_net_popup(const char * noti, const char * message)
{
	int ret = 0;
	bundle *b;
	static gboolean is_found_noti_exists = FALSE;
	static gboolean is_portal_noti_exists = FALSE;
	static gboolean is_conflict_noti_exists = FALSE;

	if (noti == NULL) {
		ERR("Invalid notification");
		return FALSE;
	}

	if (g_strcmp0(noti, NETCONFIG_DEL_FOUND_AP_NOTI) == 0) {
		if (is_found_noti_exists == FALSE)
			return TRUE;

		is_found_noti_exists = FALSE;
	} else if (g_strcmp0(noti, NETCONFIG_ADD_FOUND_AP_NOTI) == 0) {
		if (is_found_noti_exists == TRUE)
			return TRUE;

		is_found_noti_exists = TRUE;
	} else if (g_strcmp0(noti, NETCONFIG_ADD_PORTAL_NOTI) == 0) {
		if (is_portal_noti_exists == TRUE)
			return TRUE;

		is_portal_noti_exists = TRUE;
	} else if (g_strcmp0(noti, NETCONFIG_DEL_PORTAL_NOTI) == 0) {
		if (is_portal_noti_exists == FALSE)
			return TRUE;

		is_portal_noti_exists = FALSE;
	} else if (g_strcmp0(noti, NETCONFIG_DEL_IP_CONFLICT_NOTI) == 0) {
		if (is_conflict_noti_exists == FALSE)
			return TRUE;

		is_conflict_noti_exists = FALSE;
	} else if (g_strcmp0(noti, NETCONFIG_ADD_IP_CONFLICT_NOTI) == 0) {
		if (is_conflict_noti_exists == TRUE)
			return TRUE;

		is_conflict_noti_exists = TRUE;
	}

	b = bundle_create();
	bundle_add(b, "_SYSPOPUP_TYPE_", noti);

	if (message != NULL) {
		DBG("message (%s)", message);
		if (g_strcmp0(noti, NETCONFIG_ADD_IP_CONFLICT_NOTI) == 0)
			bundle_add(b, "_MAC_ADDRESS_", message);
		else
			bundle_add(b, "_AP_NAME_", message);
	}

	ret = syspopup_launch("net-popup", b);

	bundle_free(b);

	if (ret < 0) {
		ERR("Unable to launch noti-popup. Err = %d", ret);
		return FALSE;
	}

	DBG("Successfully sent notification (%s)", noti);
	return TRUE;
}

int headed_send_message_to_net_popup(const char *title,
		const char *content, const char *type, const char *ssid)
{
	int ret = 0;
	bundle *b = bundle_create();

	bundle_add(b, "_SYSPOPUP_TITLE_", title);
	bundle_add(b, "_SYSPOPUP_CONTENT_", content);
	bundle_add(b, "_SYSPOPUP_TYPE_", type);
	bundle_add(b, "_AP_NAME_", ssid);

	ret = syspopup_launch("net-popup", b);

	bundle_free(b);

	return ret;
}

int headed_send_restriction_to_net_popup(const char *title,
		const char *type, const char *restriction)
{
	int ret = 0;
	bundle *b = bundle_create();

	bundle_add(b, "_SYSPOPUP_TITLE_", title);
	bundle_add(b, "_SYSPOPUP_CONTENT_", "security restriction");
	bundle_add(b, "_SYSPOPUP_TYPE_", type);
	bundle_add(b, "_RESTRICTED_TYPE_", restriction);

	ret = syspopup_launch("net-popup", b);

	bundle_free(b);

	return ret;
}

void headed_set_system_event(int sys_evt, int evt_key, int evt_val)
{
	bundle *b = NULL;
	const char *sevent = _headed_convert_sys_event(sys_evt);
	const char *ekey = _headed_convert_sys_event(evt_key);
	const char *eval = _headed_convert_sys_event(evt_val);

	DBG("System event set [%s : %s : %s]", sevent, ekey, eval);

	b = bundle_create();
	bundle_add_str(b, ekey, eval);
	eventsystem_send_system_event(sevent, b);
	bundle_free(b);
}

void headed_pop_wifi_connected_poppup(const char *ssid)
{
	bundle *b = NULL;

	if (ssid == NULL)
		return;

	b = bundle_create();

	bundle_add(b, "_SYSPOPUP_TITLE_", "Network connection popup");
	bundle_add(b, "_SYSPOPUP_TYPE_", "toast_popup");
	bundle_add(b, "_SYSPOPUP_CONTENT_", "wifi connected");
	bundle_add(b, "_AP_NAME_", ssid);

	DBG("Launch Wi-Fi connected alert network popup");
	syspopup_launch("net-popup", b);

	bundle_free(b);
}


extern struct netconfig_headed_plugin_t netconfig_headed_plugin
		__attribute__ ((visibility("default")));
struct netconfig_headed_plugin_t netconfig_headed_plugin = {
		headed_pop_device_picker,
		headed_send_notification_to_net_popup,
		headed_send_message_to_net_popup,
		headed_send_restriction_to_net_popup,
		headed_set_system_event,
		headed_pop_wifi_connected_poppup
};

