/*
 * Network Configuration Module
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include "plugin.h"
#include "netdbus.h"
#include <stc.h>

#ifdef USE_NETCONFIG_LOG
#include "log.h"
#else
#include <dlog.h>

#define NETCONFIG_TAG		"NETCONFIG"
#define __LOG(level, format, arg...) \
	do { \
		SLOG(level, NETCONFIG_TAG, format, ## arg); \
	} while (0)

#define DBG(format, arg...)	__LOG(LOG_DEBUG, format, ## arg)
#define ERR(format, arg...)	__LOG(LOG_ERROR, format, ## arg)
#endif

#define STC_DBUS_REPLY_TIMEOUT   (120 * 1000)

#define STC_MANAGER_SERVICE                  "net.stc"
#define STC_MANAGER_STATS_PATH               "/net/stc/statistics"
#define STC_MANAGER_STATS_INTERFACE          STC_MANAGER_SERVICE ".statistics"
#define STC_MANAGER_METHOD_STATS_GET_ALL     "GetAll"

#define STC_STATS_INFO_APP_ID                "app_id"
#define STC_STATS_INFO_INCOMING_BYTES        "cnt_in_bytes"
#define STC_STATS_INFO_OUTGOING_BYTES        "cnt_out_bytes"

typedef void (*stc_dbus_dict_cb) (const char *key, GVariant *value, void *user_data);

static GVariant *__invoke_dbus_method(const char *dest, const char *path,
		const char *interface_name, const char *method, GVariant *params)
{

	GError *error = NULL;
	GVariant *reply = NULL;
	GDBusConnection *connection;

	connection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
	if (connection == NULL) {
		ERR("[STC] Failed to get GDBusconnection");
		return reply;
	}

	reply = g_dbus_connection_call_sync(
			connection,
			dest,
			path,
			interface_name,
			method,
			params,
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			STC_DBUS_REPLY_TIMEOUT,
			NULL,
			&error);

	g_object_unref(connection);

	if (reply == NULL) {
		if (error != NULL) {
			ERR("[STC] g_dbus_connection_call_sync() failed"
						"error [%d: %s]", error->code, error->message);
			g_error_free(error);
		} else {
			ERR("[STC] g_dbus_connection_call_sync() failed");
		}

		return NULL;
	}

	return reply;
}

static void __stc_extract_cb(const char *key, GVariant *value, void *user_data)
{
	battery_app_data_s *app_data = (battery_app_data_s *)user_data;

	if (app_data == NULL)
		return;

	if (!g_strcmp0(key, STC_STATS_INFO_APP_ID)) {
		const char *app_id = g_variant_get_string(value, NULL);
		app_data->app_id = g_strdup(app_id);
	} else if (!g_strcmp0(key, STC_STATS_INFO_INCOMING_BYTES)) {
		app_data->rx = g_variant_get_int64(value);
	} else if (!g_strcmp0(key, STC_STATS_INFO_OUTGOING_BYTES)) {
		app_data->tx = g_variant_get_int64(value);
	}
}

static void __stc_dbus_dictionary_foreach(GVariantIter *iter,
			stc_dbus_dict_cb cb, void *user_data)
{
	char *key = NULL;
	GVariant *value = NULL;

	if (!cb)
		return;

	while (g_variant_iter_loop(iter, "{sv}", &key, &value)) {
		if (key)
			cb(key, value, user_data);
	}
}

static void __stc_get_stats(GVariant *params, GSList **list)
{
	GVariant *message = NULL;
	GVariantIter *iter = NULL;
	GVariantIter *iter_row = NULL;
	stc_error_e dbus_result = STC_ERROR_NONE;

	message = __invoke_dbus_method(STC_MANAGER_SERVICE,
		STC_MANAGER_STATS_PATH, STC_MANAGER_STATS_INTERFACE,
		STC_MANAGER_METHOD_STATS_GET_ALL, params);

	if (message == NULL)
		return;

	g_variant_get(message, "(iaa{sv})", &dbus_result, &iter);

	while (g_variant_iter_next(iter, "a{sv}", &iter_row)) {
		battery_app_data_s *app_data = g_try_malloc0(sizeof(battery_app_data_s));
		if (app_data != NULL) {
			memset(app_data, 0, sizeof(battery_app_data_s));

			__stc_dbus_dictionary_foreach(iter_row, __stc_extract_cb, app_data);

			if (strstr(app_data->app_id, "TOTAL_") != NULL) {
				g_free(app_data->app_id);
				g_free(app_data);
			} else {
				*list = g_slist_append(*list, app_data);
			}
		}
		g_variant_iter_free(iter_row);
	}

	g_variant_iter_free(iter);
	g_variant_unref(message);
}

static void __stc_print_stats(GSList *list)
{
	GSList *stats_list = NULL;

	for (stats_list = list; stats_list != NULL; stats_list = stats_list->next) {
		battery_app_data_s *app_data = (battery_app_data_s *)stats_list->data;

		DBG("[STC] app_id[%s] rx[%d] tx[%d]", app_data->app_id, app_data->rx, app_data->tx);
	}
}

static void __stc_make_params(time_t from, time_t to, stc_iface_type_e iftype, GVariant **params)
{
	GVariantBuilder *builder;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));

	g_variant_builder_add(builder, "{sv}", "from", g_variant_new_uint64(from));
	g_variant_builder_add(builder, "{sv}", "to", g_variant_new_uint64(to));
	g_variant_builder_add(builder, "{sv}", "iftype", g_variant_new_uint16(iftype));

	*params = g_variant_new("(@a{sv})", g_variant_builder_end(builder));

	g_variant_builder_unref(builder);
}

void stc_get_dn_stats(time_t from, time_t to, GSList **list)
{
	GVariant *params = NULL;

	DBG("[STC] DN stats from [%ld] to [%ld]", from, to);

	__stc_make_params(from, to, STC_IFACE_DATACALL, &params);
	__stc_get_stats(params, list);
	__stc_print_stats(*list);
}

void stc_get_wifi_stats(time_t from, time_t to, GSList **list)
{
	GVariant *params = NULL;

	DBG("[STC] wi-fi stats from [%ld] to [%ld]", from, to);

	__stc_make_params(from, to, STC_IFACE_WIFI, &params);
	__stc_get_stats(params, list);
	__stc_print_stats(*list);
}

extern struct netconfig_stc_plugin_t netconfig_stc_plugin
		__attribute__ ((visibility("default")));
struct netconfig_stc_plugin_t netconfig_stc_plugin = {
	stc_get_dn_stats,
	stc_get_wifi_stats
};
