/*
 * Network Configuration Module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <unistd.h>
#include <stdlib.h>

#include "log.h"
#include "util.h"
#include "network-state.h"
#include "netdbus.h"
#include "clatd-handler.h"

static gboolean g_is_running = FALSE;

static void __netconfig_clatd_async_callback(GObject *source_object,
		GAsyncResult *res, gpointer user_data)
{
	GDBusConnection *conn = NULL;
	GVariant *reply = NULL;
	GError *error = NULL;
	DBG("__netconfig_clatd_async_callback");

	conn = G_DBUS_CONNECTION(source_object);
	reply = g_dbus_connection_call_finish(conn, res, &error);
	if (error) {
		ERR("Dbus error : %s", error->message);
		g_error_free(error);
	} else {
		int is_clat_enabled = -1;
		g_variant_get(reply, "(i)", &is_clat_enabled);

		if (is_clat_enabled == 0) {
			DBG("OK");
		} else if (is_clat_enabled == 1) {
			INFO("already on going");
			g_is_running = TRUE;
		} else {
			ERR("Failed to clat enable, %d", is_clat_enabled);
		}
	}

	if (reply)
		g_variant_unref(reply);
}

int netconfig_clatd_enable(void)
{
	DBG("");
	GVariant *params = NULL;
	GVariantBuilder *builder = NULL;
	gboolean rv = FALSE;
	struct clatd_ctrl_hint hint = {0 ,};
	char ns_buff[256] = {0, };

	if (g_is_running) {
		INFO("clatd is already ENABLED");
		return 0;
	}

	if (!netconfig_get_connected_cellular_internet_ipv6only_profile(&hint)) {
		ERR("Failed to get cellular profile");
		return -1;
	}

	const char *if_name = netconfig_get_default_ifname();
	if (if_name == NULL) {
		int idx;
		for (idx = 0; idx < MAX_DNS; idx++)
			g_free(hint.nameserver[idx]);

		ERR("There is no interface name");
		return -1;
	} else {
		DBG("if_name : %s", if_name);
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE ("a{is}"));

	g_variant_builder_add(builder, "{is}", 'i', if_name);

	// nameserver addresses
	if (hint.nameserver[0] != NULL || hint.nameserver[1] != NULL) {
		g_snprintf(ns_buff, 256, "%s;%s",
					((hint.nameserver[0] != NULL) ? hint.nameserver[0] : "NULL"),
					((hint.nameserver[1] != NULL) ? hint.nameserver[1] : "NULL"));
	}
	g_variant_builder_add(builder, "{is}", 'd', ns_buff);

	params = g_variant_builder_end(builder);
	g_variant_builder_unref(builder);

	rv = netconfig_invoke_dbus_method_nonblock(CLATD_SERVICE, CLATD_PATH,
			CLATD_INTERFACE, "Start", g_variant_new("(@a{is})", params), __netconfig_clatd_async_callback);

	if (!rv) {
		DBG("Failed to dbus call");
		g_variant_unref(params);

		int idx;
		for (idx = 0; idx < MAX_DNS; idx++)
			g_free(hint.nameserver[idx]);
		return -1;
	}

	g_is_running = TRUE;
	DBG("clatd enabled[%d]", g_is_running);
	int idx;
	for (idx = 0; idx < MAX_DNS; idx++)
		g_free(hint.nameserver[idx]);
	return 0;
}

int netconfig_clatd_disable(void)
{
	DBG("");
	gboolean rv = FALSE;

	if (!g_is_running) {
		ERR("clatd is already DISABLED");
		return 0;
	}

	rv = netconfig_invoke_dbus_method_nonblock(CLATD_SERVICE, CLATD_PATH,
			CLATD_INTERFACE, "Stop", NULL, NULL);
	g_is_running = FALSE;
	DBG("clatd disabled[%d]", g_is_running);
	if (!rv) {
		DBG("Failed to dbus call");
		return -1;
	}

	return 0;
}

void netconfig_clatd_reset()
{
	DBG("");
	g_is_running = FALSE;
	netconfig_clatd_enable();
}
