/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>

#include "log.h"

#include "util.h"
#include "netdbus.h"
#include "neterror.h"
#include "network-dump.h"
#include <unistd.h>

#define NETWORK_DUMP_SCRIPT  "/opt/var/lib/net-config/network_dump.sh"

#define TCP_DUMP_IN_BOOT_TIME "/opt/usr/data/tcp_dump_mode"

static gboolean tcpdump_running = FALSE;

static const char *get_current_time(void)
{
	struct timeval mytime;
	struct tm stTempTime;
	static char time_string[128];


	struct tm {
		int tm_sec; // Seconds
		int tm_min; // Minutes
		int tm_hour; // Hour (0--23)
		int tm_mday; // Day of month (1--31)
		int tm_mon; // Month (0--11)
		int tm_year; // Year (calendar year minus 1900)
		int tm_wday; // Weekday (0--6; Sunday = 0)
		int tm_yday; // Day of year (0--365)
		int tm_isdst; // 0 if daylight savings time is not in effect)
	};

	gettimeofday(&mytime, NULL);
	localtime_r(&mytime.tv_sec, &stTempTime);
	g_snprintf(time_string, 128, "%04d.%02d.%02d-%02d_%02d_%02d",
			stTempTime.tm_year + 1900, stTempTime.tm_mon + 1, stTempTime.tm_mday,
			stTempTime.tm_hour, stTempTime.tm_min, stTempTime.tm_sec);
	time_string[127] = '\0';

	return time_string;
}

static int _start_dump(gchar *dump_path)
{
	int rv = 0;
	gchar *path = NETWORK_DUMP_SCRIPT;
	char *const args[] = { NETWORK_DUMP_SCRIPT, dump_path, NULL };
	char *const envs[] = { NULL };

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0) {
		ERR("Fail to execute network log dump shell");
		return -EIO;
	}
	return 0;
}

static gboolean _stop_tcpdump(void)
{
	int ret;
	const char *path = KILLALL_EXEC_PATH;
	char *const args[] = { KILLALL_EXEC_PATH, "tcpdump", NULL };
	char *const envs[] = { NULL };

	if (!tcpdump_running) {
		DBG("tcpdump is not running");
		return FALSE;
	}

	ret = netconfig_execute_file(path, args, envs);
	if (ret < 0) {
		DBG("Failed to Kill tcpdump, ret : %d\n", ret);
		return FALSE;
	}

	tcpdump_running = FALSE;

	return TRUE;
}

static gboolean _start_tcpdump(void)
{
	int ret;
	const char *start_path = TCPDUMP_EXEC_PATH;
	char filename[128] = { 0, };
	g_snprintf(filename, 128, "%s/tcpdump-%s.pcap", NETWORK_LOG_PATH, get_current_time());
	char *const start_args[] = { TCPDUMP_EXEC_PATH,
			"-W", "5", "-C", "50", "-i", "any", "-s", "0", "-w", filename, NULL};

	if (tcpdump_running) {
		DBG("kill tcpdump.");
		gboolean ret = _stop_tcpdump();
		if (!ret)
			DBG("Failed to kill tcpdump.. ret : %d", ret);
	}

	ret = netconfig_execute_file_no_wait(start_path, start_args);
	if (ret < 0) {
		DBG("Failed to start tcpdump, ret : %d", ret);
		return FALSE;
	}

	tcpdump_running = TRUE;

	return tcpdump_running;
}

static gboolean _get_dump_state(void)
{
	int ret = access(TCP_DUMP_IN_BOOT_TIME, F_OK);
	if (ret != 0) {
		ERR("Dumpstate Get Failed (%d).", ret);
		return FALSE;
	}

	return TRUE;
}

gboolean handle_start_tcpdump(Tcpdump *object,
		GDBusMethodInvocation *context)
{
	gboolean ret = FALSE;

	ret = _start_tcpdump();
	if (!ret) {
		netconfig_error_dbus_method_return(context,
				NETCONFIG_ERROR_INTERNAL, "StartTCPDumpFailed");
		return TRUE;
	}

	DBG("Successfully started tcpdump and running..");
	tcpdump_complete_start_tcpdump(object, context);

	return TRUE;
}

gboolean handle_stop_tcpdump(Tcpdump *object, GDBusMethodInvocation *context)
{
	gboolean ret = FALSE;

	if (!tcpdump_running) {
		DBG("tcpdump not running");
		tcpdump_complete_stop_tcpdump(object, context);
		return TRUE;
	}

	ret = _stop_tcpdump();
	if (!ret) {
		netconfig_error_dbus_method_return(context,
				NETCONFIG_ERROR_INTERNAL, "StopTCPDumpFailed");
		return TRUE;
	}

	DBG("Successfully stopped tcpdump");
	tcpdump_complete_stop_tcpdump(object, context);

	return TRUE;
}

gboolean handle_get_tcpdump_state(Tcpdump *object, GDBusMethodInvocation *context)
{
	DBG("tcpdump %s running", tcpdump_running ? "is" : "is not");

	tcpdump_complete_get_tcpdump_state(object, context, tcpdump_running);

	return TRUE;
}

static void _send_dump_signal(const gchar *sig_name)
{
	gboolean reply;
	GVariant *params;
	GDBusConnection *connection = NULL;
	GError *error = NULL;

	connection = netdbus_get_connection();
	if (connection == NULL) {
		DBG("GDBusconnection is NULL");
		return;
	}

	params = g_variant_new("(i)", getpid());
	reply = g_dbus_connection_emit_signal(connection,
			NULL,
			DUMP_SERVICE_OBJECT_PATH,
			DUMP_SERVICE_INTERFACE,
			sig_name,
			params,
			&error);
	if (reply != TRUE) {
		if (error != NULL) {
			ERR("Failed to send signal [%s]", error->message);
			g_error_free(error);
		}
		return;
	}
}

int netconfig_dump_log(const char *path)
{
	gchar *dump_path = NULL;

	if (!path) {
		ERR("path is NULL. Dump Fail");
		return -1;
	}
	ERR("Dump is started");
	_send_dump_signal(DUMP_START_SIGNAL);

	dump_path = g_strdup(path);
	_start_dump(dump_path);
	_stop_tcpdump();
	g_free(dump_path);

	_send_dump_signal(DUMP_FINISH_SIGNAL);
	ERR("Dump is finished");
	return 0;
}

void check_dump_state_and_start(void)
{
	if (!_get_dump_state())
		return;

	if (!_start_tcpdump()) {
		ERR("Failed to start tcpdump");
		return;
	}
	DBG("Successfully started tcpdump and running..");

}
