/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <dlfcn.h>
#include <errno.h>
#include <vconf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <net/if.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/limits.h>
#include <ctype.h>
#include <vconf-keys.h>
#include <tzplatform_config.h>
#include <system_info.h>

#include "log.h"
#include "util.h"
#include "neterror.h"
#include "wifi-state.h"
#include "netdbus.h"

#define DBUS_SERVICE_DBUS		"org.freedesktop.DBus"
#define DBUS_INTERFACE_DBUS		"org.freedesktop.DBus"
#define MAC_INFO_FILEPATH		tzplatform_mkpath(TZ_SYS_ETC, "/.mac.info")
#define MAC_ADDRESS_FILEPATH	"/sys/class/net/wlan0/address"
#define MAC_ADDRESS_MAX_LEN		18
#define HEADED_PLUGIN_FILEPATH		"/usr/lib/net-config-plugin-headed.so"
#define TELEPHONY_PLUGIN_FILEPATH	"/usr/lib/net-config-plugin-telephony.so"
#define STC_PLUGIN_FILEPATH	        "/usr/lib/net-config-plugin-stc.so"
#define BATTERY_PLUGIN_FILEPATH	    "/usr/lib/net-config-plugin-battery.so"

static gboolean netconfig_device_picker_test = FALSE;
static int mdnsd_ref_count = 0;
typedef struct {
	char *conn_name;
	int conn_id;
} dnssd_conn_destroy_data;

static gboolean netconfig_plugin_headed_enabled = FALSE;
static gboolean netconfig_plugin_telephony_enabled = FALSE;
static gboolean netconfig_plugin_stc_enabled = FALSE;
static gboolean netconfig_plugin_battery_enabled = FALSE;
static void *handle_headed;
static void *handle_telephony;
static void *handle_stc;
static void *handle_battery;
static struct netconfig_headed_plugin_t *headed_plugin;
static struct netconfig_telephony_plugin_t *telephony_plugin;
static struct netconfig_stc_plugin_t *stc_plugin;
static struct netconfig_battery_plugin_t *battery_plugin;

static bool is_feature_checked[NETCONFIG_SUPPORTED_FEATURE_MAX] = {0, };
static bool feature_supported[NETCONFIG_SUPPORTED_FEATURE_MAX] = {0, };

gboolean netconfig_check_passphrase(const gchar *service, const char *passphrase)
{
	gsize length;

	if (!passphrase)
		return FALSE;

	length = strlen(passphrase);

	if (g_str_has_suffix(service, "psk") == TRUE) {
		if (length == 64) {
			for (int i = 0; i < 64; i++)
				if (!isxdigit((unsigned char)passphrase[i]))
					return FALSE;
		} else if (length < 8 || length > 63)
			return FALSE;
	} else if (g_str_has_suffix(service, "wep") == TRUE) {
		if (length == 10 || length == 26) {
			for (int i = 0; i < length; i++)
				if (!isxdigit((unsigned char)passphrase[i]))
					return FALSE;
		} else if (length != 5 && length != 13)
			return FALSE;
	}

	return TRUE;
}

GKeyFile *netconfig_keyfile_load(const char *pathname)
{
	GKeyFile *keyfile = NULL;
	GError *error = NULL;

	keyfile = g_key_file_new();
	if (g_key_file_load_from_file(keyfile, pathname, 0, &error) != TRUE) {
		DBG("Unable to open %s, error %s", pathname, error->message);
		g_error_free(error);

		g_key_file_free(keyfile);
		keyfile = NULL;
	}

	DBG("loaded keyfile %s", pathname);
	return keyfile;
}

void netconfig_keyfile_save(GKeyFile *keyfile, const char *pathname)
{
	gsize size = 0;
	GError *error = NULL;
	gchar *keydata = NULL;
	gchar *needle = NULL, *directory = NULL;

	directory = g_strdup(pathname);
	needle = g_strrstr(directory, "/");

	if (needle != NULL)
		*needle = '\0';

	if (directory == NULL || (*directory) == '\0') {
		g_free(directory);
		ERR("directory is NULL");
		return;
	}

	if (g_file_test(directory, G_FILE_TEST_IS_DIR) != TRUE) {
		if (g_mkdir_with_parents(directory,
				S_IRUSR | S_IWUSR | S_IXUSR) != 0) {
			g_free(directory);
			ERR("failed to make directory");
			return;
		}
	}
	g_free(directory);

	keydata = g_key_file_to_data(keyfile, &size, &error);
	if (g_file_set_contents(pathname, keydata, size, &error) != TRUE) {
		ERR("Unable to save %s, error %s", pathname, error->message);
		g_error_free(error);
	}

	if (chmod(pathname, S_IRUSR | S_IWUSR) < 0)
		DBG("Failed to change mode");
	else
		DBG("Successfully saved keyfile %s", pathname);

	g_free(keydata);
}

void netconfig_start_timer_seconds(guint secs,
		gboolean(*callback) (gpointer), void *user_data, guint *timer_id)
{
	guint t_id = 0;

	if (callback == NULL) {
		ERR("callback function is NULL");
		return;
	}

	if ((timer_id != NULL && *timer_id != 0)) {
		ERR("timer already is registered");
		return;
	}

	t_id = g_timeout_add_seconds(secs, callback, user_data);

	if (t_id == 0) {
		ERR("Can't add timer");
		return;
	}

	if (timer_id != NULL)
		*timer_id = t_id;
}

void netconfig_start_timer(guint msecs,
		gboolean(*callback) (gpointer), void *user_data, guint *timer_id)
{
	guint t_id = 0;

	INFO("Register timer with callback pointer (%p)", callback);

	if (callback == NULL) {
		ERR("callback function is NULL");
		return;
	}

	if ((timer_id != NULL && *timer_id != 0)) {
		ERR("timer already is registered");
		return;
	}

	t_id = g_timeout_add(msecs, callback, user_data);

	if (t_id == 0) {
		ERR("Can't add timer");
		return;
	}

	if (timer_id != NULL)
		*timer_id = t_id;
}

void netconfig_stop_timer(guint *timer_id)
{
	if (timer_id == NULL) {
		ERR("timer is NULL");
		return;
	}

	if (*timer_id != 0) {
		g_source_remove(*timer_id);
		*timer_id = 0;
	}
}

static gboolean __netconfig_test_device_picker()
{
	char *favorite_wifi_service = NULL;

	favorite_wifi_service = wifi_get_favorite_service();
	if (favorite_wifi_service != NULL) {
		ERR("favorite_wifi_service is existed[%s] : Donot launch device picker", favorite_wifi_service);
		g_free(favorite_wifi_service);
		return FALSE;
	}

	return TRUE;
}

static void __netconfig_pop_device_picker(void)
{
	if (!netconfig_plugin_headed_enabled)
		return;

	if (!headed_plugin)
		return;

	headed_plugin->pop_device_picker();
}

static gboolean __netconfig_wifi_try_device_picker(gpointer data)
{
	if (__netconfig_test_device_picker() == TRUE)
		__netconfig_pop_device_picker();

	return FALSE;
}

static guint __netconfig_wifi_device_picker_timer_id(gboolean is_set_method, guint timer_id)
{
	static guint netconfig_wifi_device_picker_service_timer = 0;

	if (is_set_method != TRUE)
		return netconfig_wifi_device_picker_service_timer;

	if (netconfig_wifi_device_picker_service_timer != timer_id)
		netconfig_wifi_device_picker_service_timer = timer_id;

	return netconfig_wifi_device_picker_service_timer;
}

static void __netconfig_wifi_device_picker_set_timer_id(guint timer_id)
{
	__netconfig_wifi_device_picker_timer_id(TRUE, timer_id);
}

static guint __netconfig_wifi_device_picker_get_timer_id(void)
{
	return __netconfig_wifi_device_picker_timer_id(FALSE, -1);
}

void netconfig_wifi_enable_device_picker_test(void)
{
	netconfig_device_picker_test = TRUE;
}

void netconfig_wifi_device_picker_service_start(void)
{
	const int NETCONFIG_WIFI_DEVICE_PICKER_INTERVAL = 700;
	guint timer_id = 0;

	if (netconfig_device_picker_test == TRUE)
		netconfig_device_picker_test = FALSE;
	else
		return;

	int wifi_ug_state;

	netconfig_vconf_get_int(VCONFKEY_WIFI_UG_RUN_STATE, &wifi_ug_state);
	if (wifi_ug_state == VCONFKEY_WIFI_UG_RUN_STATE_ON_FOREGROUND)
		return;

	DBG("Register device picker timer with %d milliseconds", NETCONFIG_WIFI_DEVICE_PICKER_INTERVAL);
	netconfig_start_timer(NETCONFIG_WIFI_DEVICE_PICKER_INTERVAL, __netconfig_wifi_try_device_picker, NULL, &timer_id);

	__netconfig_wifi_device_picker_set_timer_id(timer_id);
}

void netconfig_wifi_device_picker_service_stop(void)
{
	guint timer_id = 0;

	timer_id = __netconfig_wifi_device_picker_get_timer_id();
	if (timer_id == 0)
		return;

	DBG("Clear device picker timer with timer_id %d", timer_id);

	netconfig_stop_timer(&timer_id);

	__netconfig_wifi_device_picker_set_timer_id(timer_id);
}

gboolean netconfig_is_wifi_direct_on(void)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_DIRECT))
		return FALSE;

	int wifi_direct_state = 0;

	netconfig_vconf_get_int(VCONFKEY_WIFI_DIRECT_STATE, &wifi_direct_state);

	DBG("Wi-Fi direct mode %d", wifi_direct_state);
	return (wifi_direct_state != 0) ? TRUE : FALSE;
}

gboolean netconfig_is_wifi_tethering_on(void)
{
	if (netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_TETHERING)) {
		int wifi_tethering_state = 0;

		netconfig_vconf_get_int(VCONFKEY_MOBILE_HOTSPOT_MODE, &wifi_tethering_state);
		DBG("Wi-Ti tethering mode %d", wifi_tethering_state);
		if ((wifi_tethering_state & VCONFKEY_MOBILE_HOTSPOT_MODE_WIFI)
				|| (wifi_tethering_state & VCONFKEY_MOBILE_HOTSPOT_MODE_WIFI_AP)) {
			DBG("Mobile AP is on");
			return TRUE;
		}
	}

	DBG("Mobile AP is off");
	return FALSE;
}

gboolean netconfig_interface_up(const char *ifname)
{
	int fd;
	struct ifreq ifr;

	fd = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
	if (fd < 0)
		return FALSE;

	memset(&ifr, 0, sizeof(ifr));
	g_strlcpy((char *)ifr.ifr_name, ifname, sizeof(ifr.ifr_name));

	if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
		close(fd);
		return FALSE;
	}

	ifr.ifr_flags |= (IFF_UP | IFF_DYNAMIC);
	if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
		close(fd);
		return FALSE;
	}

	close(fd);

	DBG("Successfully activated wireless interface %s", ifname);
	return TRUE;
}

gboolean netconfig_interface_down(const char *ifname)
{
	int fd;
	struct ifreq ifr;

	fd = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
	if (fd < 0)
		return FALSE;

	memset(&ifr, 0, sizeof(ifr));
	g_strlcpy((char *)ifr.ifr_name, ifname, sizeof(ifr.ifr_name));

	if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
		close(fd);
		return FALSE;
	}

	ifr.ifr_flags = (ifr.ifr_flags & ~IFF_UP) | IFF_DYNAMIC;
	if (ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
		close(fd);
		return FALSE;
	}

	close(fd);

	DBG("Successfully de-activated wireless interface %s", ifname);
	return TRUE;
}

int netconfig_execute_file(const char *file_path,
		char *const args[], char *const envs[])
{
	pid_t pid = 0;
	int status = 0;
	int rv = 0;
	errno = 0;
	register unsigned int index = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	while (args[index] != NULL) {
		DBG("%s", args[index]);
		index++;
	}

	if (!(pid = fork())) {
		DBG("pid(%d), ppid (%d)", getpid(), getppid());
		DBG("Inside child, exec (%s) command", file_path);

		errno = 0;
		if (execve(file_path, args, envs) == -1) {
			DBG("Fail to execute command (%s)",
					strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
			exit(1);
		}
	} else if (pid > 0) {
		if (waitpid(pid, &status, 0) == -1)
			DBG("wait pid (%u) status (%d)", pid, status);

		if (WIFEXITED(status)) {
			rv = WEXITSTATUS(status);
			DBG("exited, status=%d", rv);
		} else if (WIFSIGNALED(status)) {
			DBG("killed by signal %d", WTERMSIG(status));
		} else if (WIFSTOPPED(status)) {
			DBG("stopped by signal %d", WSTOPSIG(status));
		} else if (WIFCONTINUED(status)) {
			DBG("continued");
		}

		return rv;
	}

	DBG("failed to fork(%s)",
		strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
	return -EIO;
}

int netconfig_execute_cmd(const char *cmd)
{
	if (cmd == NULL)
		return -EIO;

	pid_t pid = 0;
	int status = 0;
	int rv = 0;
	errno = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };
	gchar **args = NULL;

	DBG("command: %s", cmd);

	args = g_strsplit_set(cmd, " ", -1);

	if (!(pid = fork())) {
		DBG("pid(%d), ppid (%d)", getpid(), getppid());

		errno = 0;
		if (execv(args[0], args) == -1) {
			DBG("Fail to execute command (%s)",
				strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
			g_strfreev(args);
			exit(1);
		}
	} else if (pid > 0) {
		if (waitpid(pid, &status, 0) == -1)
			DBG("wait pid (%u) status (%d)", pid, status);

		if (WIFEXITED(status)) {
			rv = WEXITSTATUS(status);
			DBG("exited, status=%d", rv);
		} else if (WIFSIGNALED(status)) {
			DBG("killed by signal %d", WTERMSIG(status));
		} else if (WIFSTOPPED(status)) {
			DBG("stopped by signal %d", WSTOPSIG(status));
		} else if (WIFCONTINUED(status)) {
			DBG("continued");
		}

		g_strfreev(args);
		return rv;
	}

	DBG("failed to fork(%s)",
		strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
	g_strfreev(args);

	return -EIO;
}

static void on_clat_handler()
{
	pid_t clat_pid = 0;
	int state = 0;

	clat_pid = waitpid(-1, &state, WNOHANG);

	DBG("clat(%d) state(%d)", clat_pid, WEXITSTATUS(state));
}

int netconfig_execute_clatd(const char *file_path, char *const args[])
{
	pid_t pid = 0;
	int rv = 0;
	errno = 0;
	register unsigned int index = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	struct sigaction act;
	int state = 0;

	act.sa_handler = on_clat_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;

	state = sigaction(SIGCHLD, &act, 0);
	if (state != 0) {
		DBG("sigaction() : %d", state);
		return -1;
	}

	while (args[index] != NULL) {
		DBG("%s", args[index]);
		index++;
	}

	if (!(pid = fork())) {
		DBG("pid(%d), ppid (%d)", getpid(), getppid());
		DBG("Inside child, exec (%s) command", file_path);

		errno = 0;
		if (execvp(file_path, args) == -1) {
			ERR("Fail to execute command (%s)",
				strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
			return -1;
		}
	} else if (pid > 0) {
		ERR("Success to launch clatd");
		return rv;
	}

	DBG("failed to fork(%s)",
		strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
	return -EIO;
}

static void no_wait_signal_handler()
{
	pid_t child_pid = 0;
	int state = 0;

	child_pid = waitpid(-1, &state, WNOHANG);

	DBG("child_id(%d) state(%d)", child_pid, WEXITSTATUS(state));
}

int netconfig_execute_file_no_wait(const char *file_path, char *const args[])
{
	pid_t pid = 0;
	int rv = 0;
	errno = 0;
	register unsigned int index = 0;

	struct sigaction act;
	int state = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	act.sa_handler = no_wait_signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;

	state = sigaction(SIGCHLD, &act, 0);
	if (state != 0) {
		DBG("sigaction() : %d", state);
		return -1;
	}

	while (args[index] != NULL) {
		DBG("%s", args[index]);
		index++;
	}

	if (!(pid = fork())) {
		DBG("pid(%d), ppid (%d)", getpid(), getppid());
		DBG("Inside child, exec (%s) command", file_path);

		errno = 0;
		if (execvp(file_path, args) == -1) {
			ERR("Fail to execute command (%s)",
				strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
			return -1;
		}
	} else if (pid > 0) {
		ERR("Successfully launched child process");
		return rv;
	}

	DBG("failed to fork(%s)",
		strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
	return -EIO;
}

int __netconfig_get_interface_index(const char *interface_name)
{
	struct ifreq ifr;
	int sock = 0;
	int result = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	if (interface_name == NULL) {
		DBG("Inteface name is NULL");
		return -1;
	}

	errno = 0;
	sock = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
	if (sock < 0) {
		DBG("Failed to create socket : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, interface_name, sizeof(ifr.ifr_name) - 1);
	result = ioctl(sock, SIOCGIFINDEX, &ifr);
	close(sock);

	if (result < 0) {
		DBG("Failed to get ifr index: %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	return ifr.ifr_ifindex;
}

int netconfig_add_route_ipv4(gchar *ip_addr, gchar *subnet, gchar *interface, gint address_family)
{
	struct ifreq ifr;
	struct rtentry rt;
	struct sockaddr_in addr_in;
	int sock;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	memset(&ifr, 0, sizeof(ifr));

	ifr.ifr_ifindex = __netconfig_get_interface_index(interface);

	if (ifr.ifr_ifindex < 0)
		return -1;

	strncpy(ifr.ifr_name, interface, IFNAMSIZ-1);

	memset(&rt, 0, sizeof(rt));

	rt.rt_flags = RTF_UP | RTF_HOST;
	memset(&addr_in, 0, sizeof(struct sockaddr_in));
	addr_in.sin_family = address_family;
	addr_in.sin_addr.s_addr = inet_addr(ip_addr);
	memcpy(&rt.rt_dst, &addr_in, sizeof(rt.rt_dst));

	memset(&addr_in, 0, sizeof(struct sockaddr_in));
	addr_in.sin_family = address_family;
	addr_in.sin_addr.s_addr = INADDR_ANY;
	memcpy(&rt.rt_gateway, &addr_in, sizeof(rt.rt_gateway));

	memset(&addr_in, 0, sizeof(struct sockaddr_in));
	addr_in.sin_family = AF_INET;
	addr_in.sin_addr.s_addr = inet_addr(subnet);
	memcpy(&rt.rt_genmask, &addr_in, sizeof(rt.rt_genmask));

	rt.rt_dev = ifr.ifr_name;

	errno = 0;
	sock = socket(PF_INET, SOCK_DGRAM, 0);

	if (sock < 0) {
		DBG("Failed to create socket : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	if (ioctl(sock, SIOCADDRT, &rt) < 0) {
		DBG("Failed to set route address : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		close(sock);
		return -1;
	}

	close(sock);

	return 1;
}

int netconfig_del_route_ipv4(gchar *ip_addr, gchar *subnet, gchar *interface, gint address_family)
{
	struct ifreq ifr;
	struct rtentry rt;
	struct sockaddr_in addr_in;
	int sock;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	memset(&ifr, 0, sizeof(ifr));
	ifr.ifr_ifindex = __netconfig_get_interface_index(interface);

	if (ifr.ifr_ifindex < 0)
		return -1;

	strncpy(ifr.ifr_name, interface, IFNAMSIZ-1);

	memset(&rt, 0, sizeof(rt));

	rt.rt_flags = RTF_UP;
	memset(&addr_in, 0, sizeof(struct sockaddr_in));
	addr_in.sin_family = address_family;
	addr_in.sin_addr.s_addr = inet_addr(ip_addr);
	memcpy(&rt.rt_dst, &addr_in, sizeof(rt.rt_dst));

	memset(&addr_in, 0, sizeof(struct sockaddr_in));
	addr_in.sin_family = address_family;
	addr_in.sin_addr.s_addr = inet_addr(subnet);
	memcpy(&rt.rt_genmask, &addr_in, sizeof(rt.rt_genmask));
	rt.rt_dev = ifr.ifr_name;

	errno = 0;
	sock = socket(PF_INET, SOCK_DGRAM, 0);

	if (sock < 0) {
		DBG("Failed to create socket : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	if (ioctl(sock, SIOCDELRT, &rt) < 0) {
		DBG("Failed to set route address : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		close(sock);
		return -1;
	}

	close(sock);

	return 1;
}

int netconfig_add_route_ipv6(gchar *ip_addr, gchar *interface, gchar *gateway, unsigned char prefix_len)
{
	struct in6_rtmsg rt;
	int fd = 0;
	int err = 0;
	char error_buf[MAX_SIZE_ERROR_BUFFER] = {0, };

	memset(&rt, 0, sizeof(rt));

	rt.rtmsg_dst_len = prefix_len;

	rt.rtmsg_flags = RTF_UP | RTF_HOST;

	errno = 0;
	if (inet_pton(AF_INET6, ip_addr, &rt.rtmsg_dst) < 0) {
		DBG("inet_pton failed : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	if (gateway != NULL) {
		rt.rtmsg_flags |= RTF_GATEWAY;
		if (inet_pton(AF_INET6, gateway, &rt.rtmsg_gateway) < 0) {
			DBG("inet_pton failed : %s",
				strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
			return -1;
		}
	}

	rt.rtmsg_metric = 1;

	fd = socket(AF_INET6, SOCK_DGRAM, 0);
	if (fd < 0) {
		DBG("Failed to create socket : %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		return -1;
	}

	rt.rtmsg_ifindex = 0;

	if (interface) {
		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
		strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name)-1);
		ioctl(fd, SIOCGIFINDEX, &ifr);
		rt.rtmsg_ifindex = ifr.ifr_ifindex;
	}

	if ((err = ioctl(fd, SIOCADDRT, &rt)) < 0) {
		DBG("Failed to add route: %s",
			strerror_r(errno, error_buf, MAX_SIZE_ERROR_BUFFER));
		close(fd);
		return -1;
	}

	close(fd);

	return 1;
}

int netconfig_del_route_ipv6(gchar *ip_addr, gchar *interface, gchar *gateway, unsigned char prefix_len)
{
	struct in6_rtmsg rt;
	int fd = 0;
	int err = 0;

	memset(&rt, 0, sizeof(rt));

	rt.rtmsg_dst_len = prefix_len;

	rt.rtmsg_flags = RTF_UP | RTF_HOST;

	if (inet_pton(AF_INET6, ip_addr, &rt.rtmsg_dst) < 0) {
		err = -errno;
		return err;
	}

	if (gateway != NULL) {
		rt.rtmsg_flags |= RTF_GATEWAY;
		if (inet_pton(AF_INET6, gateway, &rt.rtmsg_gateway) < 0) {
			err = -errno;
			return err;
		}
	}

	rt.rtmsg_metric = 1;

	fd = socket(AF_INET6, SOCK_DGRAM, 0);
	if (fd < 0)
		return -1;

	rt.rtmsg_ifindex = 0;

	if (interface) {
		struct ifreq ifr;
		memset(&ifr, 0, sizeof(ifr));
		strncpy(ifr.ifr_name, interface, sizeof(ifr.ifr_name)-1);
		ioctl(fd, SIOCGIFINDEX, &ifr);
		rt.rtmsg_ifindex = ifr.ifr_ifindex;
	}

	if ((err = ioctl(fd, SIOCDELRT, &rt)) < 0) {
		DBG("Failed to del route: %d\n", err);
		close(fd);
		return -1;
	}

	close(fd);

	return 1;
}

gboolean handle_launch_direct(Wifi *wifi, GDBusMethodInvocation *context)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_DIRECT)) {
		wifi_complete_launch_direct(wifi, context);
		return TRUE;
	}

	int ret = 0;
	DBG("Launch Wi-Fi direct daemon");

	const char *path = "/usr/bin/wifi-direct-server.sh";
	char *const args[] = { "wifi-direct-server.sh", "start", NULL };
	char *const envs[] = { NULL };

	ret = netconfig_execute_file(path, args, envs);
	if (ret < 0) {
		ERR("Failed to launch Wi-Fi direct daemon");
		netconfig_error_wifi_direct_failed(context);
		return TRUE;
	}

	wifi_complete_launch_direct(wifi, context);
	return TRUE;
}

int execute_mdnsd_script(char* op)
{
	const char *path = "/usr/bin/mdnsresponder-server.sh";
	char *const args[] = { "mdnsresponder-server.sh", op, NULL };
	char *const envs[] = { NULL };

	return netconfig_execute_file(path, args, envs);
}

static void __dnssd_conn_destroyed_cb(GDBusConnection *conn,
		const gchar *Name, const gchar *path, const gchar *interface,
		const gchar *sig, GVariant *param, gpointer user_data)
{
	gchar *name = NULL;
	gchar *old = NULL;
	gchar *new = NULL;
	dnssd_conn_destroy_data *data = user_data;
	GDBusConnection *connection = NULL;
	connection = netdbus_get_connection();

	if (param == NULL)
		return;

	g_variant_get(param, "(sss)", &name, &old, &new);

	if (g_strcmp0(name, data->conn_name) == 0 && *new == '\0') {
		DBG("Connection %s Destroyed: name %s id %d", data->conn_name, name,
			data->conn_id);
		mdnsd_ref_count--;
		g_dbus_connection_signal_unsubscribe(connection, data->conn_id);
		if (mdnsd_ref_count == 0) {
			if (execute_mdnsd_script("stop") < 0)
				ERR("Failed to stop mdnsresponder daemon");
		}
	}
	g_free(name);
	g_free(old);
	g_free(new);
	g_free(data->conn_name);
	g_free(data);
	return;
}

static void register_dnssd_conn_destroy_signal(gchar *name)
{
	dnssd_conn_destroy_data *data;
	GDBusConnection *connection = NULL;
	connection = netdbus_get_connection();

	if (connection == NULL) {
		ERR("Failed to get GDbus Connection");
		return;
	}

	data = g_try_malloc0(sizeof(dnssd_conn_destroy_data));

	if (data == NULL) {
		ERR("Out of Memory!");
		return;
	}

	data->conn_name = g_strdup(name);

	data->conn_id = g_dbus_connection_signal_subscribe(connection,
							DBUS_SERVICE_DBUS, DBUS_INTERFACE_DBUS,
							"NameOwnerChanged", NULL, name,
							G_DBUS_SIGNAL_FLAGS_NONE, __dnssd_conn_destroyed_cb,
							data, NULL);
	return;
}

gboolean handle_launch_mdns(Network *object, GDBusMethodInvocation *context,
							gchar *name)
{
	DBG("Launch mdnsresponder daemon");

	if (execute_mdnsd_script("start") < 0) {
		ERR("Failed to launch mdnsresponder daemon");
		netconfig_error_invalid_parameter(context);
		return TRUE;
	}

	mdnsd_ref_count++;
	register_dnssd_conn_destroy_signal(name);
	DBG("Ref mdnsresponder daemon. ref count: %d", mdnsd_ref_count);

	network_complete_launch_mdns(object, context);
	return TRUE;
}

gboolean netconfig_send_notification_to_net_popup(const char * noti, const char * ssid)
{
	if (!netconfig_plugin_headed_enabled)
		return FALSE;

	if (!headed_plugin)
		return FALSE;

	return headed_plugin->send_notification_to_net_popup(noti, ssid);
}

int netconfig_send_message_to_net_popup(const char *title,
		const char *content, const char *type, const char *ssid)
{
	if (!netconfig_plugin_headed_enabled)
		return 0;

	if (!headed_plugin)
		return 0;

	return headed_plugin->send_message_to_net_popup(title, content, type, ssid);
}

int netconfig_send_restriction_to_net_popup(const char *title,
		const char *type, const char *restriction)
{
	if (!netconfig_plugin_headed_enabled)
		return 0;

	if (!headed_plugin)
		return 0;

	return headed_plugin->send_restriction_to_net_popup(title, type, restriction);
}

void netconfig_set_system_event(int sys_evt, int evt_key, int evt_val)
{
	if (!netconfig_plugin_headed_enabled)
		return;

	if (!headed_plugin)
		return;

	headed_plugin->set_system_event(sys_evt, evt_key, evt_val);
}

void __netconfig_pop_wifi_connected_poppup(const char *ssid)
{
	if (!netconfig_plugin_headed_enabled)
		return;

	if (!headed_plugin)
		return;

	headed_plugin->pop_wifi_connected_poppup(ssid);
}

void netconfig_get_telephony_network_type(int *svctype, int *pstype)
{
	if (!netconfig_plugin_telephony_enabled)
		return;

	if (!telephony_plugin)
		return;

	telephony_plugin->get_telephony_network_type(svctype, pstype);
}

gboolean __netconfig_wifi_get_sim_imsi(Wifi *wifi, GDBusMethodInvocation *context)
{
	if (!netconfig_plugin_telephony_enabled)
		return FALSE;

	if (!telephony_plugin)
		return FALSE;

	return telephony_plugin->wifi_get_sim_imsi(wifi, context);
}

netconfig_error_e __netconfig_wifi_req_aka_auth(GArray *rand_data, GArray *autn_data,
		GDBusMethodInvocation *context, struct wifi_authentication_data **data)
{
	if (!netconfig_plugin_telephony_enabled)
		return NETCONFIG_ERROR_INTERNAL;

	if (!telephony_plugin)
		return NETCONFIG_ERROR_INTERNAL;

	return telephony_plugin->wifi_req_aka_auth(rand_data, autn_data, context, data);
}

gboolean __netconfig_wifi_req_sim_auth(GArray *rand_data,
		GDBusMethodInvocation *context, struct wifi_authentication_data **data)
{
	if (!netconfig_plugin_telephony_enabled)
		return FALSE;

	if (!telephony_plugin)
		return FALSE;

	return telephony_plugin->wifi_req_sim_auth(rand_data, context, data);
}

gboolean netconfig_tapi_check_sim_state(void)
{
	if (!netconfig_plugin_telephony_enabled)
		return FALSE;

	if (!telephony_plugin)
		return FALSE;

	return telephony_plugin->tapi_check_sim_state();
}

gboolean __netconfig_wifi_get_aka_authdata(Wifi *wifi,
		GDBusMethodInvocation *context, struct wifi_authentication_data **data)
{
	if (!netconfig_plugin_telephony_enabled)
		return FALSE;

	if (!telephony_plugin)
		return FALSE;

	return telephony_plugin->wifi_get_aka_authdata(wifi, context, data);
}

gboolean __netconfig_wifi_get_sim_authdata(Wifi *wifi,
		GDBusMethodInvocation *context, struct wifi_authentication_data **data)
{
	if (!netconfig_plugin_telephony_enabled)
		return FALSE;

	if (!telephony_plugin)
		return FALSE;

	return telephony_plugin->wifi_get_sim_authdata(wifi, context, data);
}

static void __netconfig_stc_get_dn_stats(time_t from, time_t to, GSList **list)
{
	if (!netconfig_plugin_stc_enabled)
		return;

	if (!stc_plugin)
		return;

	return stc_plugin->get_stc_dn_stats(from, to, list);
}

static void __netconfig_stc_get_wifi_stats(time_t from, time_t to, GSList **list)
{
	if (!netconfig_plugin_stc_enabled)
		return;

	if (!stc_plugin)
		return;

	return stc_plugin->get_stc_wifi_stats(from, to, list);
}

void netconfig_battery_start_dn(void)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->start_dn_data();
}

void netconfig_battery_end_dn(void)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->end_dn_data(0, __netconfig_stc_get_dn_stats);
}

void netconfig_battery_update_dn_rssi(int rssi)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->update_dn_rssi(rssi);
}

void netconfig_battery_start_wifi(void)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->start_wifi_data();
}

void netconfig_battery_end_wifi(void)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->end_wifi_data(0, __netconfig_stc_get_wifi_stats);
}

void netconfig_battery_update_wifi_scan(int state)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->update_wifi_scan(state);
}

void netconfig_battery_update_wifi_rssi(int rssi)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->update_wifi_rssi(rssi);
}

void netconfig_battery_get_dn_list(void *data)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->get_battery_dn_list(data, __netconfig_stc_get_dn_stats);
}

void netconfig_battery_get_wifi_list(void *data)
{
	if (!netconfig_plugin_battery_enabled)
		return;

	if (!battery_plugin)
		return;

	return battery_plugin->get_battery_wifi_list(data, __netconfig_stc_get_wifi_stats);
}

void netconfig_set_vconf_int(const char * key, int value)
{
	int ret = 0;

	DBG("[%s: %d]", key, value);

	ret = vconf_set_int(key, value);
	if (ret != VCONF_OK)
		ERR("Failed to set");
}

void netconfig_set_vconf_str(const char * key, const char * value)
{
	int ret = 0;

	DBG("[%s: %s]", key, value);

	ret = vconf_set_str(key, value);
	if (ret != VCONF_OK)
		ERR("Failed to set");
}

int netconfig_vconf_get_int(const char * key, int *value)
{
	int ret = 0;

	ret = vconf_get_int(key, value);
	if (ret != VCONF_OK) {
		ERR("Failed to get vconfkey [%s] value", key);
		return -1;
	}

	return 0;
}

int netconfig_vconf_get_bool(const char * key, int *value)
{
	int ret = 0;

	ret = vconf_get_bool(key, value);
	if (ret != VCONF_OK) {
		ERR("Failed to get vconfkey [%s] value", key);
		return -1;
	}

	return 0;
}

char* netconfig_get_env(const char *key)
{
	FILE *fp;
	char buf[256], *entry = NULL, *value = NULL, *last;
	int len = 0;

	if (!key)
		return NULL;

	fp = fopen(NETCONFIG_TIZEN_SYSTEM_ENV, "r");
	if (!fp)
		return NULL;

	while (fgets(buf, sizeof(buf), fp)) {
		entry = buf;
		entry = strtok_r(entry, "=", &last);
		if (entry) {
			if (strstr(entry, key)) {
				entry = strtok_r(NULL, "\n", &last);
				if (entry) {
					len = strlen(entry);
					value = (char*)malloc(len+1);
					g_strlcpy(value, entry, len+1);
				} else {
					value = (char*)malloc(sizeof(char));
					g_strlcpy(value, "\n", sizeof(char));
				}
				break;
			}
		}
	}

	fclose(fp);
	return value;
}

void netconfig_set_mac_address_from_file(void)
{
	FILE *file = NULL;
	char mac_str[MAC_ADDRESS_MAX_LEN];
	gchar *mac_lower_str = NULL;
	int mac_len = 0;

	file = fopen(MAC_INFO_FILEPATH, "r");
	if (file == NULL) {
		ERR("Fail to open %s", MAC_INFO_FILEPATH);
		file = fopen(MAC_ADDRESS_FILEPATH, "r");
		if (file == NULL) {
			ERR("Fail to open %s", MAC_ADDRESS_FILEPATH);
			return;
		}
	}
	if (fgets(mac_str, sizeof(mac_str), file) == NULL) {
		ERR("Fail to read mac address");
		fclose(file);
		return;
	}

	mac_len = strlen(mac_str);
	if (mac_len < 17) {
		ERR("mac.info is empty");
		fclose(file);
		return;
	}

	mac_lower_str = g_ascii_strup(mac_str, (gssize)mac_len);
	netconfig_set_vconf_str(VCONFKEY_WIFI_BSSID_ADDRESS, mac_lower_str);

	g_free(mac_lower_str);
	fclose(file);
}

char *netconfig_get_mac_address_from_file(const char *ifname)
{
	FILE *file = NULL;
	char file_path[PATH_MAX] = {0, };
	char mac_str[MAC_ADDRESS_MAX_LEN];
	int mac_len = 0;

	if (!ifname)
		return NULL;

	g_snprintf(file_path, sizeof(file_path),
			"/sys/class/net/%s/address", ifname);

	DBG("ifname: %s", ifname);
	DBG("file_path: %s", file_path);
	file = fopen(file_path, "r");
	if (file == NULL) {
		ERR("Fail to open %s", file_path);
		return NULL;
	}
	if (fgets(mac_str, sizeof(mac_str), file) == NULL) {
		ERR("Fail to read mac address");
		fclose(file);
		return NULL;
	}

	mac_len = strlen(mac_str);
	if (mac_len < 17) {
		ERR("mac is empty");
		fclose(file);
		return NULL;
	}
	DBG("address: %s", mac_str);

	fclose(file);
	return g_strdup(mac_str);
}

int netconfig_freq_to_channel(int freq)
{
	if (freq < 2412 || freq > 5825 ||
		(freq > 2484 && freq < 5180)) {
		ERR("Invalid Frequence Range");
		return 0;
	}
	if (freq >= 5180)
		return 36 + (freq - 5180)/5;
	else if (freq <= 2472)
		return 1 + (freq - 2412)/5;
	else if (freq == 2484)
		return 14;
	else
		return 0;
}

int netconfig_get_operating_class(int freq)
{
	int channel = 0;
	int oper_class = 0;

	channel = netconfig_freq_to_channel(freq);

	if (channel) {
		/* Operating class 81 - 2.4 GHz band channels 1..13 */
		if (channel >= 1 && channel <= 13)
			oper_class = 81;
		/* Operating class 115 - 5 GHz, channels 36-48 */
		else if (channel >= 36 && channel <= 48)
			oper_class = 115;
		/* Operating class 124 - 5 GHz, channels 149,153,157,161 */
		else
			oper_class = 124;

		INFO("Operating Class  is [%d]", oper_class);
		return oper_class;
	}
	return 0;
}

tizen_profile_t _get_tizen_profile()
{
	static tizen_profile_t profile = TIZEN_PROFILE_UNKNOWN;
	if (__builtin_expect(profile != TIZEN_PROFILE_UNKNOWN, 1))
		return profile;

	char *profileName;
	system_info_get_platform_string("http://tizen.org/feature/profile", &profileName);
	switch (*profileName) {
	case 'm':
	case 'M':
		profile = TIZEN_PROFILE_MOBILE;
		break;
	case 'w':
	case 'W':
		profile = TIZEN_PROFILE_WEARABLE;
		break;
	case 't':
	case 'T':
		profile = TIZEN_PROFILE_TV;
		break;
	case 'i':
	case 'I':
		profile = TIZEN_PROFILE_IVI;
		break;
	default: // common or unknown ==> ALL ARE COMMON.
		profile = TIZEN_PROFILE_COMMON;
	}
	free(profileName);

	return profile;
}

void netconfig_plugin_init()
{
	handle_headed = dlopen(HEADED_PLUGIN_FILEPATH, RTLD_NOW);
	if (!handle_headed) {
		ERR("Can't load %s: %s", HEADED_PLUGIN_FILEPATH, dlerror());
	} else {
		headed_plugin = dlsym(handle_headed, "netconfig_headed_plugin");
		if (!headed_plugin) {
			ERR("Can't load symbol: %s", dlerror());
			dlclose(handle_headed);
		} else {
			netconfig_plugin_headed_enabled = TRUE;
		}
	}

	handle_telephony = dlopen(TELEPHONY_PLUGIN_FILEPATH, RTLD_NOW);
	if (!handle_telephony) {
		ERR("Can't load %s: %s", TELEPHONY_PLUGIN_FILEPATH, dlerror());
	} else {
		telephony_plugin = dlsym(handle_telephony, "netconfig_telephony_plugin");
		if (!telephony_plugin) {
			ERR("Can't load symbol: %s", dlerror());
			dlclose(handle_telephony);
		} else {
			netconfig_plugin_telephony_enabled = TRUE;
		}
	}

	handle_stc = dlopen(STC_PLUGIN_FILEPATH, RTLD_NOW);
	if (!handle_stc) {
		ERR("Can't load %s: %s", STC_PLUGIN_FILEPATH, dlerror());
	} else {
		stc_plugin = dlsym(handle_stc, "netconfig_stc_plugin");
		if (!stc_plugin) {
			ERR("Can't load symbol: %s", dlerror());
			dlclose(handle_stc);
		} else {
			netconfig_plugin_stc_enabled = TRUE;
		}
	}

	handle_battery = dlopen(BATTERY_PLUGIN_FILEPATH, RTLD_NOW);
	if (!handle_battery) {
		ERR("Can't load %s: %s", BATTERY_PLUGIN_FILEPATH, dlerror());
	} else {
		battery_plugin = dlsym(handle_battery, "netconfig_battery_plugin");
		if (!battery_plugin) {
			ERR("Can't load symbol: %s", dlerror());
			dlclose(handle_battery);
		} else {
			netconfig_plugin_battery_enabled = TRUE;
		}
	}

}

void netconfig_plugin_deinit()
{
	if (netconfig_plugin_headed_enabled) {
		netconfig_plugin_headed_enabled = FALSE;
		dlclose(handle_headed);
	}

	if (netconfig_plugin_telephony_enabled) {
		netconfig_plugin_telephony_enabled = FALSE;
		dlclose(handle_telephony);
	}

	if (netconfig_plugin_stc_enabled) {
		netconfig_plugin_stc_enabled = FALSE;
		dlclose(handle_stc);
	}

	if (netconfig_plugin_battery_enabled) {
		netconfig_plugin_battery_enabled = FALSE;
		dlclose(handle_battery);
	}

}

gboolean netconfig_get_headed_plugin_flag()
{
	return netconfig_plugin_headed_enabled;
}

gboolean netconfig_get_telephony_plugin_flag()
{
	return netconfig_plugin_telephony_enabled;
}

bool netconfig_check_feature_supported(netconfig_supported_feature_e feature)
{
	const char *key = NULL;

	if (!is_feature_checked[feature]) {
		switch (feature) {
		case NETCONFIG_SUPPORTED_FEATURE_ETHERNET:
			key = ETHERNET_FEATURE;
			break;
		case NETCONFIG_SUPPORTED_FEATURE_TETHERING:
			key = TETHERING_FEATURE;
			break;
		case NETCONFIG_SUPPORTED_FEATURE_WIFI_DIRECT:
			key = WIFI_DIRECT_FEATURE;
			break;
		case NETCONFIG_SUPPORTED_FEATURE_WIFI_SOFTAP:
			key = WIFI_SOFTAP_FEATURE;
			break;
		default:
			ERR("Uknown feature");
			return false;
		}

		if (system_info_get_platform_bool(key, &feature_supported[feature]) < 0) {
			ERR("Get feature is failed");
			return false;
		}
		is_feature_checked[feature] = true;
	}
	return feature_supported[feature];
}

void netconfig_convert_bytes_to_hexstr(const char *bin, int blen, gchar* hexstr)
{
	char t;

	while (blen) {
		t = (*bin >> 4) & 0x0f;

		if (t <= 9)
			*hexstr = t + '0';
		else if (t >= 10 && t <= 16)
			*hexstr = (t - 10) + 'a';

		hexstr++;

		t = *bin & 0x0f;

		if (t <= 9)
			*hexstr = t + '0';
		else if (t >= 10 && t <= 16)
			*hexstr = (t - 10) + 'a';

		hexstr++;
		bin++;
		blen--;
	}

	*hexstr = '\0';
}
