/*
 * Network Configuration - VPN Service Module
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <gio/gunixfdlist.h>

#include "vpnsvc.h"
#include "vpnsvc-internal.h"
#include "netdbus.h"
#include "log.h"

static Vpnsvc *vpnsvc = NULL;

/*********************
 * Handler Functions *
 ********************/
gboolean handle_vpn_init(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								const gchar *arg_iface_name,
								guint arg_iface_name_len)
{
	DBG("handle_vpn_init");

	int result = VPNSVC_ERROR_NONE;

	vpnsvc_tun_s handle_s;
	GDBusMessage *msg;
	GUnixFDList *fd_list;
	int fd_list_length;
	const int *fds;

	DBG("vpn_init, %s, %u\n", arg_iface_name, arg_iface_name_len);

	msg = g_dbus_method_invocation_get_message(invocation);
	fd_list = g_dbus_message_get_unix_fd_list(msg);
	fds = g_unix_fd_list_peek_fds(fd_list, &fd_list_length);

	if (fd_list_length <= 0)
		DBG("D-Bus Message doesn't contain any fd!");

	DBG("fd:%d\n", *fds);

	result = vpn_service_init(arg_iface_name, arg_iface_name_len, *fds, &handle_s);

	DBG("handle_s.fd : %d, handle_s.index : %d, handle_s.name : %s",
			handle_s.fd, handle_s.index, handle_s.name);

	vpnsvc_complete_vpn_init(object, invocation, result, handle_s.index, handle_s.name);

	return TRUE;
}

gboolean handle_vpn_deinit(Vpnsvc *object,
									GDBusMethodInvocation *invocation,
									const gchar *arg_dev_name)
{
	DBG("handle_vpn_deinit");

	int result = VPNSVC_ERROR_NONE;

	DBG("vpn_deinit, %s\n", arg_dev_name);

	result = vpn_service_deinit(arg_dev_name);

	vpnsvc_complete_vpn_deinit(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_protect(Vpnsvc *object,
									GDBusMethodInvocation *invocation,
									const gchar *arg_dev_name)
{
	DBG("handle_vpn_protect");

	int result = VPNSVC_ERROR_NONE;

	int socket;
	GDBusMessage *msg;
	GUnixFDList *fd_list;
	int fd_list_length;
	const int *fds;

	msg = g_dbus_method_invocation_get_message(invocation);
	fd_list = g_dbus_message_get_unix_fd_list(msg);
	fds = g_unix_fd_list_peek_fds(fd_list, &fd_list_length);
	if (fd_list_length <= 0)
		DBG("D-Bus Message doesn't contain any fd!");

	socket = *fds;
	DBG("vpn_protect, %d, %s\n", socket, arg_dev_name);

	result = vpn_service_protect(socket, arg_dev_name);

	vpnsvc_complete_vpn_protect(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_up(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								const gchar *arg_iface_name)
{
	DBG("handle_vpn_up");

	int result = VPNSVC_ERROR_NONE;

	result = vpn_service_up(arg_iface_name);

	vpnsvc_complete_vpn_up(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_down(Vpnsvc *object,
									GDBusMethodInvocation *invocation,
									const gchar *arg_iface_name)
{
	DBG("handle_vpn_down");

	int result = VPNSVC_ERROR_NONE;

	result = vpn_service_down(arg_iface_name);

	vpnsvc_complete_vpn_down(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_block_networks(Vpnsvc *object,
											GDBusMethodInvocation *invocation,
											GVariant *arg_nets_vpn,
											guint arg_nr_nets_vpn,
											GVariant *arg_nets_orig,
											guint arg_nr_nets_orig)
{
	DBG("handle_vpn_block_networks");

	int result = VPNSVC_ERROR_NONE;

	char *nets_vpn[arg_nr_nets_vpn];
	int prefix_vpn[arg_nr_nets_vpn];

	char *nets_orig[arg_nr_nets_vpn];
	int prefix_orig[arg_nr_nets_vpn];

	int i = 0;
	GVariantIter iter;
	gchar* route_dest;
	gint route_prefix;

	DBG("vpn_block_networks");

	memset(nets_vpn, 0, sizeof(char *) * arg_nr_nets_vpn);
	memset(prefix_vpn, 0, sizeof(int) * arg_nr_nets_vpn);
	memset(nets_orig, 0, sizeof(char *) * arg_nr_nets_vpn);
	memset(prefix_orig, 0, sizeof(int) * arg_nr_nets_vpn);

	/* arg_nets_vpn check */
	if (arg_nr_nets_vpn > 0) {
		if (arg_nets_vpn != NULL) {
			GVariant *dict_nets_vpn = g_variant_get_variant(arg_nets_vpn);
			g_variant_iter_init(&iter, dict_nets_vpn);
			i = 0;
			while (g_variant_iter_loop(&iter, "{si}", &route_dest, &route_prefix)) {
				int tmp_route_len = strlen(route_dest);
				nets_vpn[i] = g_try_malloc0(sizeof(char) * tmp_route_len + 1);
				strncpy(nets_vpn[i], route_dest, tmp_route_len);
				nets_vpn[i][tmp_route_len] = '\0';
				prefix_vpn[i] = route_prefix;
				DBG("nets_vpn[%d] = %s \t", i, (nets_vpn[i] == NULL) ? "" : nets_vpn[i]);
				DBG("prefix_vpn[%d] = %d ", i, prefix_vpn[i]);
				i++;
			}
		}
	}

	/* arg_nets_orig check */
	if (arg_nr_nets_orig > 0) {
		if (arg_nets_orig != NULL) {
			GVariant *dict_nets_orig = g_variant_get_variant(arg_nets_orig);
			g_variant_iter_init(&iter, dict_nets_orig);
			i = 0;
			while (g_variant_iter_loop(&iter, "{si}", &route_dest, &route_prefix)) {
				int tmp_route_len = strlen(route_dest);
				nets_orig[i] = g_try_malloc0(sizeof(char) * tmp_route_len + 1);
				strncpy(nets_orig[i], route_dest, tmp_route_len);
				nets_orig[i][tmp_route_len] = '\0';
				prefix_orig[i] = route_prefix;
				DBG("nets_orig[%d] = %s \t", i, (nets_orig[i] == NULL) ? "" : nets_orig[i]);
				DBG("prefix_orig[%d] = %d ", i, prefix_orig[i]);
				i++;
			}
		}
	}

	/* call function */
	result = vpn_service_block_networks(nets_vpn, prefix_vpn, arg_nr_nets_vpn, nets_orig, prefix_orig, arg_nr_nets_orig);

	for (i = 0; i < arg_nr_nets_vpn; ++i) {
		g_free(nets_orig[i]);
		g_free(nets_vpn[i]);
	}

	vpnsvc_complete_vpn_block_networks(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_unblock_networks(Vpnsvc *object,
											GDBusMethodInvocation *invocation)
{
	DBG("handle_vpn_unblock_networks");

	int result = VPNSVC_ERROR_NONE;

	DBG("vpn_unblock_networks");

	result = vpn_service_unblock_networks();

	vpnsvc_complete_vpn_unblock_networks(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_update_settings(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								gint arg_iface_index,
								const gchar *arg_local_ip,
								const gchar *arg_remote_ip,
								guint arg_mtu)
{
	int result = VPNSVC_ERROR_NONE;
	DBG("handle_vpn_update_settings");

	result = vpn_service_update_settings(arg_iface_index, arg_local_ip, arg_remote_ip, arg_mtu);

	vpnsvc_complete_vpn_update_settings(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_add_route(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								gchar *arg_iface_name,
								const gchar *arg_route,
								gint arg_prefix)
{
	DBG("handle_vpn_add_route");

	int result = VPNSVC_ERROR_NONE;

	result = vpn_service_add_route(arg_iface_name, arg_route, arg_prefix);

	vpnsvc_complete_vpn_add_route(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_remove_route(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								gchar *arg_iface_name,
								const gchar *arg_route,
								gint arg_prefix)

{
	DBG("handle_vpn_remove_route");

	int result = VPNSVC_ERROR_NONE;

	result = vpn_service_remove_route(arg_iface_name, arg_route, arg_prefix);

	vpnsvc_complete_vpn_remove_route(object, invocation, result);

	return TRUE;
}

gboolean handle_vpn_add_dns_server(Vpnsvc *object,
								GDBusMethodInvocation *invocation,
								gchar *arg_iface_name,
								const gchar *arg_dns_server)
{
	DBG("handle_vpn_add_dns_server");

	int result = VPNSVC_ERROR_NONE;

	result = vpn_service_add_dns_server(arg_iface_name, arg_dns_server);

	vpnsvc_complete_vpn_add_dns_server(object, invocation, result);

	return TRUE;
}

/*****************************
 * Initializations Functions *
 ****************************/
Vpnsvc *get_vpnsvc_object(void)
{
	return vpnsvc;
}

void vpnsvc_create_and_init(void)
{
	DBG("Create vpn object.");
	GDBusInterfaceSkeleton *interface_vpn = NULL;
	GDBusConnection *connection = NULL;
	GDBusObjectManagerServer *server = netdbus_get_vpn_manager();
	if (server == NULL)
		return;

	connection = netdbus_get_connection();
	g_dbus_object_manager_server_set_connection(server, connection);

	/* Interface */
	vpnsvc = vpnsvc_skeleton_new();
	interface_vpn = G_DBUS_INTERFACE_SKELETON(vpnsvc);

	/* VPN Service */
	g_signal_connect(vpnsvc, "handle-vpn-init",
			G_CALLBACK(handle_vpn_init), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-deinit",
			G_CALLBACK(handle_vpn_deinit), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-protect",
			G_CALLBACK(handle_vpn_protect), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-up",
			G_CALLBACK(handle_vpn_up), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-down",
			G_CALLBACK(handle_vpn_down), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-block-networks",
			G_CALLBACK(handle_vpn_block_networks), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-unblock-networks",
			G_CALLBACK(handle_vpn_unblock_networks), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-update-settings",
			G_CALLBACK(handle_vpn_update_settings), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-add-route",
			G_CALLBACK(handle_vpn_add_route), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-remove-route",
			G_CALLBACK(handle_vpn_remove_route), NULL);
	g_signal_connect(vpnsvc, "handle-vpn-add-dns-server",
			G_CALLBACK(handle_vpn_add_dns_server), NULL);

	if (!g_dbus_interface_skeleton_export(interface_vpn, connection,
			NETCONFIG_VPNSVC_PATH, NULL)) {
		ERR("Export NETCONFIG_VPNSVC_PATH for vpn failed");
	}

	return;
}

void vpnsvc_destroy_deinit(void)
{
	DBG("Deinit vpn object.");

	if (vpnsvc)
		g_object_unref(vpnsvc);
}

