/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <errno.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <stdio.h>
#include <fcntl.h>

#include "log.h"
#include "util.h"
#include "netdbus.h"
#include "emulator.h"
#include "neterror.h"
#include "netsupplicant.h"
#include "wifi-firmware.h"
#include "network-statistics.h"

#define WLAN_DRIVER_SCRIPT			"/usr/bin/wlan.sh"
#define WLAN_IFACE_NAME				"wlan0"

#define WLAN_P2P_IFACE_NAME_TV			"p2p0"
#define WLAN_P2P_IFACE_NAME_COMMON			"wlan0"
#define WLAN_P2P_IFACE_NAME ((TIZEN_TV) ? (WLAN_P2P_IFACE_NAME_TV) : (WLAN_P2P_IFACE_NAME_COMMON))
#define QUAD_CPUS_COUNT			4
#define TEMP_BUFFER_LEN			100
#define WIFI_MAC_ADD_PATH		"/sys/class/net/wlan0/address"

static int __netconfig_sta_firmware_start(void)
{
	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "start", NULL };
	char *const envs[] = { NULL };

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

	rv = netconfig_interface_up(WLAN_IFACE_NAME);
	if (rv != TRUE)
		return -EIO;

	DBG("Successfully loaded wireless device driver");
	return 0;
}

static int __netconfig_sta_firmware_stop(void)
{
	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "stop", NULL };
	char *const envs[] = { NULL };

	/* Update statistics before driver remove */
	netconfig_wifi_statistics_update_powered_off();

	rv = netconfig_interface_down(WLAN_IFACE_NAME);
	if (rv != TRUE)
		return -EIO;

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

	DBG("Successfully removed wireless device driver");
	return 0;
}

static int __netconfig_p2p_firmware_start(void)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_DIRECT))
		return -ENODEV;

	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "p2p", NULL };
	char *const envs[] = { NULL };

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

#if defined TIZEN_WLAN_USE_P2P_INTERFACE
	rv = netconfig_interface_up(WLAN_P2P_IFACE_NAME);
	if (rv != TRUE)
		return -EIO;
#endif

	DBG("Successfully loaded p2p device driver");
	return 0;
}

static int __netconfig_p2p_firmware_stop(void)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_DIRECT))
		return -ENODEV;

	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "stop", NULL };
	char *const envs[] = { NULL };

	rv = netconfig_interface_down(WLAN_IFACE_NAME);
	if (rv != TRUE)
		return -EIO;

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

	DBG("Successfully removed p2p device driver");
	return 0;
}

static int __netconfig_softap_firmware_start(void)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_TETHERING)
			&& !netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_SOFTAP))
		return -ENODEV;

	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "softap", NULL };
	char *const envs[] = { NULL };

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

	if (netconfig_interface_up(WLAN_IFACE_NAME) == FALSE)
		return -EIO;

	DBG("Successfully loaded softap device driver");
	return 0;
}

static int __netconfig_softap_firmware_stop(void)
{
	if (!netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_TETHERING)
			&& !netconfig_check_feature_supported(NETCONFIG_SUPPORTED_FEATURE_WIFI_SOFTAP))
		return -ENODEV;

	int rv = 0;
	const char *path = WLAN_DRIVER_SCRIPT;
	char *const args[] = { "/usr/bin/wlan.sh", "stop", NULL };
	char *const envs[] = { NULL };

	rv = netconfig_interface_down(WLAN_IFACE_NAME);
	if (rv != TRUE)
		return -EIO;

	rv = netconfig_execute_file(path, args, envs);
	if (rv < 0)
		return -EIO;

	DBG("Successfully removed softap device driver");
	return 0;
}

static int __netconfig_wifi_firmware_start(enum netconfig_wifi_firmware type)
{
	if (emulator_is_emulated() == TRUE)
		return -EIO;

	switch (type) {
	case NETCONFIG_WIFI_STA:
		return __netconfig_sta_firmware_start();
	case NETCONFIG_WIFI_P2P:
		return __netconfig_p2p_firmware_start();
	case NETCONFIG_WIFI_SOFTAP:
		return __netconfig_softap_firmware_start();
	default:
		break;
	}

	return -ENXIO;
}

static int __netconfig_wifi_firmware_stop(enum netconfig_wifi_firmware type)
{
	if (emulator_is_emulated() == TRUE)
		return -EIO;

	switch (type) {
	case NETCONFIG_WIFI_STA:
		return __netconfig_sta_firmware_stop();
	case NETCONFIG_WIFI_P2P:
		return __netconfig_p2p_firmware_stop();
	case NETCONFIG_WIFI_SOFTAP:
		return __netconfig_softap_firmware_stop();
	default:
		break;
	}

	return -ENXIO;
}

static int __netconfig_set_rps_cpus(void)
{
	int fd, curr;
	ssize_t count;
	char t_buf[TEMP_BUFFER_LEN];
	char r_buf[TEMP_BUFFER_LEN];

	if (access(WIFI_MAC_ADD_PATH, F_OK) != 0) {
		DBG("WiFi driver is not loaded... ");
		return -1;
	} else {
		DBG("WiFi driver loaded... ");
	}

	snprintf(t_buf, TEMP_BUFFER_LEN, "/sys/class/net/wlan0/queues/rx-0/rps_cpus");
	DBG("Command : [%s]", t_buf);
	curr = 0;

	while ((fd = open(t_buf, O_RDWR | O_CLOEXEC)) >= 0) {
		curr++;
		count = read(fd, r_buf, 1);

		if (count < 0) {
			DBG("read failed");
			close(fd);
			return -1;
		} else {
			DBG("read size = %zd", count);
		}

		if (r_buf[0] == 'e') {
			close(fd);
			DBG("e is already written");
			snprintf(t_buf, TEMP_BUFFER_LEN, "/sys/class/net/wlan0/queues/rx-%d/rps_cpus", curr);
			DBG("Command : [%s]", t_buf);
			continue;
		} else {
			DBG("writing e");
		}

		if (lseek(fd, 0, SEEK_SET) < 0) {
			DBG("lseek failed");
			close(fd);
			return -1;
		}

		count = write(fd, "e", 1);

		if (count < 0) {
			DBG("write failed");
			close(fd);
			return -1;
		} else {
			DBG("write size = %zd", count);
		}

		close(fd);
		snprintf(t_buf, TEMP_BUFFER_LEN, "/sys/class/net/wlan0/queues/rx-%d/rps_cpus", curr);
		DBG("Command : [%s]", t_buf);
	}

	return 0;
}

int netconfig_wifi_firmware(enum netconfig_wifi_firmware type, gboolean enable)
{
	int err;
	static enum netconfig_wifi_firmware current_driver = NETCONFIG_WIFI_OFF;
	enum netconfig_wifi_firmware alias = type;

	DBG("Wi-Fi current firmware %d (type: %d %s)", current_driver, type,
							enable == TRUE ? "enable" : "disable");

	if (enable == FALSE) {
		if (current_driver == NETCONFIG_WIFI_OFF) {
			return -EALREADY;
		} else if (current_driver == alias) {

			err = __netconfig_wifi_firmware_stop(type);
			if (err < 0 && err != -EALREADY)
				return err;

			current_driver = NETCONFIG_WIFI_OFF;

			return err;
		}

		return -EIO;
	}

	if (current_driver > NETCONFIG_WIFI_OFF) {
		if (current_driver == alias)
			return -EALREADY;

		return -EIO;
	}

	err = __netconfig_wifi_firmware_start(type);
	if (err < 0)
		DBG("Failed to execute script file");
	else
		current_driver = alias;

	if (__netconfig_set_rps_cpus() < 0)
		DBG("Failed to set rps_cpus");

	return err;
}

gboolean handle_start(WifiFirmware *firmware, GDBusMethodInvocation *context, const gchar *device)
{
	int err;

	g_return_val_if_fail(firmware != NULL, TRUE);

	DBG("Wi-Fi firmware start %s", device != NULL ? device : "null");

	if (g_strcmp0("p2p", device) == 0)
		err = netconfig_wifi_firmware(NETCONFIG_WIFI_P2P, TRUE);
	else if (g_strcmp0("softap", device) == 0)
		err = netconfig_wifi_firmware(NETCONFIG_WIFI_SOFTAP, TRUE);
	else
		err = -EINVAL;

	if (err < 0) {
		if (err == -EALREADY)
			netconfig_error_already_exists(context);
		else if (g_strcmp0("softap", device) == 0 && err == -EIO && netconfig_is_wifi_direct_on() == FALSE) {
			if (netconfig_wifi_firmware(NETCONFIG_WIFI_P2P, FALSE) == 0 && netconfig_wifi_firmware(NETCONFIG_WIFI_SOFTAP, TRUE) == 0) {
				wifi_firmware_complete_start(firmware, context);
				return TRUE;
			} else
				netconfig_error_wifi_driver_failed(context);
		} else
			netconfig_error_wifi_driver_failed(context);

		return TRUE;
	}

	wifi_firmware_complete_start(firmware, context);
	return TRUE;
}

gboolean handle_stop(WifiFirmware *firmware, GDBusMethodInvocation *context, const gchar *device)
{
	int err;

	g_return_val_if_fail(firmware != NULL, TRUE);

	DBG("Wi-Fi firmware stop %s", device != NULL ? device : "null");

	if (g_strcmp0("p2p", device) == 0)
		err = netconfig_wifi_firmware(NETCONFIG_WIFI_P2P, FALSE);
	else if (g_strcmp0("softap", device) == 0)
		err = netconfig_wifi_firmware(NETCONFIG_WIFI_SOFTAP, FALSE);
	else
		err = -EINVAL;

	if (err < 0) {
		if (err == -EALREADY)
			netconfig_error_already_exists(context);
		else
			netconfig_error_wifi_driver_failed(context);

		return TRUE;
	}

	wifi_firmware_complete_stop(firmware, context);
	return TRUE;
}
