/*
 * Network Configuration Module
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>

#include "netsupplicant.h"
#include "log.h"
#include "util.h"
#include "wifi-config.h"
#include "wifi-netlink-scan.h"
#include <netlink/genl/genl.h>
#include <netlink/genl/family.h>
#include <netlink/genl/ctrl.h>
#include <netlink/msg.h>
#include <netlink/attr.h>
#include <netlink/netlink.h>
#include <ctype.h>

static GSList *bss_info_list = NULL;

void __netconfig_notify_netlink_scan_done(void)
{
	GVariantBuilder *builder = NULL;
	GVariantBuilder *builder1 = NULL;
	GVariantBuilder *builder2 = NULL;
	GSList* list = NULL;
	const char *prop_ssid = "ssid";
	const char *prop_bssid = "bssid";
	const char *prop_freq = "freq";
	const char *prop_rssi = "rssi";
	const char *prop_vsie_list = "vsie_list";
	const char *prop_sec = "security";
	const char *prop_enc = "encryption";

	builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
	for (list = bss_info_list; list != NULL; list = list->next) {
		struct bss_scan_info_t *bss_info = (struct bss_scan_info_t *)list->data;

		if (bss_info) {
			char *bssid = (char *)bss_info->bssid;
			char *ssid = (char *)bss_info->ssid;
			GSList *vsie_list = bss_info->vsie_list;
			int freq = (int)bss_info->freq;
			int signal = (int)bss_info->signal;
			int sec_type = (int)bss_info->security_type;
			int enc_type = (int)bss_info->encryption_type;

			g_variant_builder_add(builder, "{sv}", prop_ssid, g_variant_new_string(ssid));
			g_variant_builder_add(builder, "{sv}", prop_bssid, g_variant_new_string(bssid));
			g_variant_builder_add(builder, "{sv}", prop_freq, g_variant_new_int32(freq));
			g_variant_builder_add(builder, "{sv}", prop_rssi, g_variant_new_int32(signal));
			builder1 = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
			GSList *list;
			unsigned char *net_vsie;
			unsigned int net_vsie_len;
			int count;
			for (list = vsie_list; list; list = list->next) {
				builder2 = g_variant_builder_new(G_VARIANT_TYPE("ay"));
				net_vsie = (unsigned char *)list->data;
				net_vsie_len = net_vsie[1] + 2;

				for (count = 0; count < net_vsie_len; count++) {
					g_variant_builder_add(builder2, "y", net_vsie[count]);
				}

				g_variant_builder_add(builder1, "{sv}", "Vsie", g_variant_builder_end(builder2));
				g_variant_builder_unref(builder2);
			}
			g_variant_builder_add(builder, "{sv}", prop_vsie_list, g_variant_builder_end(builder1));
			g_variant_builder_unref(builder1);

			if (vsie_list != NULL)
				g_slist_free_full(vsie_list, g_free);

			vsie_list = NULL;

			g_variant_builder_add(builder, "{sv}", prop_sec, g_variant_new_int32(sec_type));
			g_variant_builder_add(builder, "{sv}", prop_enc, g_variant_new_int32(enc_type));
		}
	}

	wifi_emit_netlink_scan_completed((Wifi *)get_wifi_object(), g_variant_builder_end(builder));
	g_variant_builder_unref(builder);

	if (bss_info_list != NULL)
		g_slist_free_full(bss_info_list, g_free);

	bss_info_list = NULL;
	INFO("NetlinkScanCompleted");

	return;
}

static int ack_handler(struct nl_msg *msg, void *user_data)
{
	int *err = user_data;
	*err = 0;
	DBG("");
	return NL_STOP;
}

static int finish_handler(struct nl_msg *msg, void *user_data)
{
	int *ret = user_data;
	*ret = 0;
	return NL_SKIP;
}

static int error_handler(struct sockaddr_nl *nla, struct nlmsgerr *err,
			 void *user_data)
{
	int *ret = user_data;
	*ret = err->error;
	return NL_SKIP;
}

static int no_seq_check(struct nl_msg *msg, void *user_data)
{
	DBG("");
	return NL_OK;
}

static int __netconfig_family_handler(struct nl_msg *msg, void *user_data)
{
	/** Callback for NL_CB_VALID in multicast group */
	struct netconfig_netlink_scan_handler_args *grp = user_data;
	struct nlattr *tb[CTRL_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct nlattr *mc_grp;
	int rem_mc_grp;

	nla_parse(tb, CTRL_ATTR_MAX, genlmsg_attrdata(gnlh, 0), genlmsg_attrlen(gnlh, 0), NULL);

	if (!tb[CTRL_ATTR_MCAST_GROUPS])
		return NL_SKIP;

	nla_for_each_nested(mc_grp, tb[CTRL_ATTR_MCAST_GROUPS], rem_mc_grp) {
		struct nlattr *tb_mc_grp[CTRL_ATTR_MCAST_GRP_MAX + 1];

		nla_parse(tb_mc_grp, CTRL_ATTR_MCAST_GRP_MAX, nla_data(mc_grp), nla_len(mc_grp), NULL);

		if (!tb_mc_grp[CTRL_ATTR_MCAST_GRP_NAME] || !tb_mc_grp[CTRL_ATTR_MCAST_GRP_ID])
			continue;
		if (strncmp(nla_data(tb_mc_grp[CTRL_ATTR_MCAST_GRP_NAME]), grp->group,
					nla_len(tb_mc_grp[CTRL_ATTR_MCAST_GRP_NAME])))
			continue;

		grp->id = nla_get_u32(tb_mc_grp[CTRL_ATTR_MCAST_GRP_ID]);
		break;
	}

	return NL_SKIP;
}

static int __netconfig_get_multicast_id(struct nl_sock *socket, const char *family, const char *group)
{
	struct nl_msg *msg = NULL;
	struct nl_cb *cb = NULL;
	int ret, ctrl_id;
	struct netconfig_netlink_scan_handler_args grp = { .group = group, .id = -ENOENT, };

	msg = nlmsg_alloc();
	if (!msg)
		return -ENOMEM;

	cb = nl_cb_alloc(NL_CB_DEFAULT);
	if (!cb) {
		ret = -ENOMEM;
		goto cb_fail;
	}

	ctrl_id = genl_ctrl_resolve(socket, "nlctrl");

	genlmsg_put(msg, 0, 0, ctrl_id, 0, 0, CTRL_CMD_GETFAMILY, 0);

	ret = -ENOBUFS;
	NLA_PUT_STRING(msg, CTRL_ATTR_FAMILY_NAME, family);

	ret = nl_send_auto_complete(socket, msg);
	if (ret < 0)
		goto out;

	ret = 1;

	nl_cb_err(cb, NL_CB_CUSTOM, error_handler, &ret);
	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &ret);
	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, __netconfig_family_handler, &grp);

	while (ret > 0)
		nl_recvmsgs(socket, cb);

	if (ret == 0)
		ret = grp.id;

nla_put_failure:
out:
	nl_cb_put(cb);
cb_fail:
	nlmsg_free(msg);
	return ret;
}

static void __netconfig_macaddress_str(char *bssid, unsigned char *user_data)
{
	int i;

	for (i = 0; i < 6; i++) {
		if (i == 0) {
			snprintf(bssid, 3, "%02x", user_data[i]);
			bssid += 2;
		} else {
			snprintf(bssid, 4, ":%02x", user_data[i]);
			bssid += 3;
		}
	}
}

typedef enum {
	WIFI_SECURITY_TYPE_NONE = 0,
	WIFI_SECURITY_TYPE_WEP = 1,
	WIFI_SECURITY_TYPE_WPA_PSK = 2,
	WIFI_SECURITY_TYPE_WPA2_PSK = 3,
	WIFI_SECURITY_TYPE_EAP = 4,
} wifi_security_type_e;

typedef enum {
	WIFI_ENCRYPTION_TYPE_NONE = 0,
	WIFI_ENCRYPTION_TYPE_WEP = 1,
	WIFI_ENCRYPTION_TYPE_TKIP = 2,
	WIFI_ENCRYPTION_TYPE_AES = 3,
	WIFI_ENCRYPTION_TYPE_TKIP_AES_MIXED = 4,
} wifi_encryption_type_e;

static unsigned char ms_oui[3]      = { 0x00, 0x50, 0xf2 };
static unsigned char ieee80211_oui[3]   = { 0x00, 0x0f, 0xac };

static void __netconfig_get_security(unsigned char *bss_element, int length, wifi_security_type_e *sec_type, wifi_encryption_type_e *enc_type)
{
	int i;
	unsigned char *data;
	uint8_t *t_data;
	int len;
	__u16 count;

	*sec_type = WIFI_SECURITY_TYPE_NONE;
	*enc_type = WIFI_ENCRYPTION_TYPE_NONE;

	while (length >= 2 && length >= bss_element[1]) {
		if (bss_element[0] == 221 && (bss_element[1] >= 4 && memcmp(bss_element + 2, ms_oui, 3) == 0)) {
			data = bss_element + 2 + 4 + 2 + 4;
			len = bss_element[1] - 4 - 2 - 4;

			if (len < 2) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			count = data[0] | (data[1] << 8);

			if (2 + (count * 4) > len) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			for (i = 0; i < count; i++) {
				t_data = (data + 2 + (i * 4));

				if ((memcmp(t_data, ms_oui, 3) == 0) || (memcmp(t_data, ieee80211_oui, 3) == 0)) {
					if (t_data[3] == 1 || t_data[3] == 5) { // 1 : WEP-40, 5 : WEP-104
						*sec_type = WIFI_SECURITY_TYPE_WEP;
						*enc_type = WIFI_ENCRYPTION_TYPE_WEP;
						return;
					} else if (t_data[3] == 2) { // 2 : TKIP
						if (*sec_type != WIFI_SECURITY_TYPE_WPA2_PSK)
							*sec_type = WIFI_SECURITY_TYPE_WPA_PSK;
						if (*enc_type == WIFI_ENCRYPTION_TYPE_AES)
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP_AES_MIXED;
						else
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP;
					} else if (t_data[3] == 4) { // 3 : CCMP
						if (*sec_type != WIFI_SECURITY_TYPE_WPA2_PSK)
							*sec_type = WIFI_SECURITY_TYPE_WPA_PSK;
						if (*enc_type == WIFI_ENCRYPTION_TYPE_TKIP)
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP_AES_MIXED;
						else
							*enc_type = WIFI_ENCRYPTION_TYPE_AES;
					}
				} //if
			} //for

			data += 2 + (count * 4);
			len -= 2 + (count * 4);

			if (len < 2) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			count = data[0] | (data[1] << 8);

			if (2 + (count * 4) > len) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			for (i = 0; i < count; i++) {
				t_data = (data + 2 + (i * 4));

				if ((memcmp(t_data, ms_oui, 3) == 0) || (memcmp(t_data, ieee80211_oui, 3) == 0)) {
					if (t_data[3] == 1 || t_data[3] == 3 || t_data[3] == 5) { // 1 : IEEE 802.1X, 3 : FT/IEEE 802.1X, 5 : IEEE 802.1X/SHA-256
						*sec_type = WIFI_SECURITY_TYPE_EAP;
					} else if (t_data[3] == 2 || t_data[3] == 4 || t_data[3] == 6) {  // 2 : PSK, 4 : FT/PSK, 6 : PSK/SHA-256
						if (*sec_type != WIFI_SECURITY_TYPE_WPA2_PSK)
							*sec_type = WIFI_SECURITY_TYPE_WPA_PSK;
					}
				}
			}
		} else if (bss_element[0] == 48 && (bss_element[1] >= 2 && bss_element[1] <= 255)) {
			data = bss_element + 2 + 2 + 4;
			len = bss_element[1] - 2 - 4;

			if (len < 2) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			count = data[0] | (data[1] << 8);
			if (2 + (count + 4) > len) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			for (i = 0; i < count; i++) {
				t_data = (data + 2 + (i * 4));

				if ((memcmp(t_data, ms_oui, 3) == 0) || (memcmp(t_data, ieee80211_oui, 3) == 0)) {
					if (t_data[3] == 1 || t_data[3] == 5) { // 1 : WEP-40, 5 : WEP-104
						*sec_type = WIFI_SECURITY_TYPE_WEP;
						*enc_type = WIFI_ENCRYPTION_TYPE_WEP;
						return;
					} else if (t_data[3] == 2) { // 2 : TKIP
						*sec_type = WIFI_SECURITY_TYPE_WPA2_PSK;
						if (*enc_type == WIFI_ENCRYPTION_TYPE_AES)
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP_AES_MIXED;
						else
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP;
					} else if (t_data[3] == 4) { // 3 : CCMP
						*sec_type = WIFI_SECURITY_TYPE_WPA2_PSK;
						if (*enc_type == WIFI_ENCRYPTION_TYPE_TKIP)
							*enc_type = WIFI_ENCRYPTION_TYPE_TKIP_AES_MIXED;
						else
							*enc_type = WIFI_ENCRYPTION_TYPE_AES;
					}
				}
			}

			data += 2 + (count * 4);
			len -= 2 + (count * 4);

			if (len < 2) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			count = data[0] | (data[1] << 8);
			if (2 + (count * 4) > len) {
				length -= bss_element[1] + 2;
				bss_element += bss_element[1] + 2;
				continue;
			}

			for (i = 0; i < count; i++) {
				t_data = (data + 2 + (i * 4));

				if ((memcmp(t_data, ms_oui, 3) == 0) || (memcmp(t_data, ieee80211_oui, 3) == 0)) {
					if (t_data[3] == 1 || t_data[3] == 3 || t_data[3] == 5) { // 1 : IEEE 802.1X, 3 : FT/IEEE 802.1X, 5 : IEEE 802.1X/SHA-256
						*sec_type = WIFI_SECURITY_TYPE_EAP;
					} else if (t_data[3] == 2 || t_data[3] == 4 || t_data[3] == 6) {  // 2 : PSK, 4 : FT/PSK, 6 : PSK/SHA-256
						*sec_type = WIFI_SECURITY_TYPE_WPA2_PSK;
					}
				}
			} //for
		} //else if

		length -= bss_element[1] + 2;
		bss_element += bss_element[1] + 2;
	} //while

	return;
}

static void __netconfig_get_vsie(unsigned char *bss_element, int length, GSList **dst)
{
	unsigned char *vsie;
	int vsie_len = 0;

	if (length < 3) {
		DBG("Vendor specific data not available");
		return;
	}

	while (length >= 2 && length >= bss_element[1]) {
		if (bss_element[0] == NETCONFIG_VENDOR_SPECIFIC_ID) {
			vsie_len = bss_element[1]+2;
			vsie = (unsigned char *)g_try_malloc0(vsie_len);

			if (vsie) {
				memcpy(vsie, bss_element, vsie_len);
				*dst = g_slist_append(*dst, vsie);
			} else
				DBG("Failed to allocate memory");
		}

		length -= bss_element[1] + 2;
		bss_element += bss_element[1] + 2;
	}
}

static void __netconfig_found_ap(unsigned char *bss_element, int length, char *str)
{
	uint8_t len;
	uint8_t *data;
	int i;

	while (length >= 2 && length >= bss_element[1]) {
		if (bss_element[0] == 0 && bss_element[1] <= 32) {
			len = bss_element[1];
			data = bss_element + 2;
			for (i = 0; i < len; i++) {
				if (isprint(data[i]) && data[i] != ' ' && data[i] != '\\')
					snprintf(&str[i], 2, "%c", data[i]);
				else if (data[i] == ' ' && (i != 0 && i != len -1))
					snprintf(&str[i], 2, "%c", ' ');
				else
					snprintf(&str[i], 2, "%c", data[i]);
			}
			break;
		}
		length -= bss_element[1] + 2;
		bss_element += bss_element[1] + 2;
	}
}

static int __netconfig_netlink_scan_cb(struct nl_msg *msg, void *user_data)
{
	/** Called by the kernel with a dump of the successful scan's data. Called for each SSID. */
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	char bssid[NETCONFIG_BSSID_LEN+1];
	char ssid[NETCONFIG_SSID_LEN+1] = {0, };
	wifi_security_type_e sec_type = WIFI_SECURITY_TYPE_NONE;
	wifi_encryption_type_e enc_type = WIFI_ENCRYPTION_TYPE_NONE;
	int wep_check = 0;
	GSList *vsie = NULL;
	struct nlattr *tb[NL80211_ATTR_MAX + 1];
	struct nlattr *bss[NL80211_BSS_MAX + 1];
	struct nla_policy bss_policy[NL80211_BSS_MAX + 1] = {
		[NL80211_BSS_FREQUENCY] = {.type = NLA_U32},
		[NL80211_BSS_BSSID] = { },
		[NL80211_BSS_CAPABILITY] = {.type = NLA_U16},
		[NL80211_BSS_INFORMATION_ELEMENTS] = { },
		[NL80211_BSS_SIGNAL_MBM] = {.type = NLA_U32},
	};

	/** Parse nl message and check error. */
	nla_parse(tb, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0), genlmsg_attrlen(gnlh, 0), NULL);
	if (!tb[NL80211_ATTR_BSS]) {
		DBG("BSS info not available");
		return NL_SKIP;
	}
	if (nla_parse_nested(bss, NL80211_BSS_MAX, tb[NL80211_ATTR_BSS], bss_policy)) {
		DBG("Failed to parse nested attributes");
		return NL_SKIP;
	}
	if ((!bss[NL80211_BSS_BSSID]) || (!bss[NL80211_BSS_INFORMATION_ELEMENTS]))
		return NL_SKIP;

	/** Extract BSSID and AP info. */
	__netconfig_macaddress_str(bssid, nla_data(bss[NL80211_BSS_BSSID]));
	__netconfig_found_ap(nla_data(bss[NL80211_BSS_INFORMATION_ELEMENTS]), nla_len(bss[NL80211_BSS_INFORMATION_ELEMENTS]), ssid);
	__netconfig_get_vsie(nla_data(bss[NL80211_BSS_INFORMATION_ELEMENTS]), nla_len(bss[NL80211_BSS_INFORMATION_ELEMENTS]), &vsie);

	if (bss[NL80211_BSS_CAPABILITY]) {
		__u16 capa = nla_get_u16(bss[NL80211_BSS_CAPABILITY]);
		if (capa & WLAN_CAPABILITY_PRIVACY) {
			wep_check = 1;
		}
	}
	__netconfig_get_security(nla_data(bss[NL80211_BSS_INFORMATION_ELEMENTS]),
			nla_len(bss[NL80211_BSS_INFORMATION_ELEMENTS]), &sec_type, &enc_type);

	if (wep_check && sec_type == WIFI_SECURITY_TYPE_NONE) {
		sec_type = WIFI_SECURITY_TYPE_WEP;
		enc_type = WIFI_ENCRYPTION_TYPE_WEP;
	}

	if (sec_type == WIFI_SECURITY_TYPE_EAP) {
		g_slist_free_full(vsie, g_free);
		return NL_SKIP;
	}
	for (GSList *list = bss_info_list; list != NULL; list = list->next) {
		struct bss_scan_info_t *bss_info = (struct bss_scan_info_t *)list->data;
		if ((g_strcmp0(bss_info->ssid, ssid) == 0) && (bss_info->security_type == sec_type)
				&& (bss_info->encryption_type == enc_type)) {
			g_slist_free_full(vsie, g_free);
			return NL_SKIP;
		}
	}

	/** Create AP info list. */
	if (ssid[0] != '\0') {
		struct bss_scan_info_t *bss_info;
		int signal;

		bss_info = g_try_new0(struct bss_scan_info_t, 1);
		if (bss_info == NULL) {
			g_slist_free_full(vsie, g_free);
			return NL_SKIP;
		}

		g_strlcpy(bss_info->bssid, bssid, strlen(bssid)+1);
		g_strlcpy(bss_info->ssid, ssid, strlen(ssid)+1);
		bss_info->vsie_list = vsie;
		bss_info->freq = nla_get_u32(bss[NL80211_BSS_FREQUENCY]);

		if (bss[NL80211_BSS_SIGNAL_MBM]) {
			signal = nla_get_u32(bss[NL80211_BSS_SIGNAL_MBM]);
			signal /= 100; /** mBm to dBm */
			bss_info->signal = signal;
		}

		bss_info->security_type = sec_type;
		bss_info->encryption_type = enc_type;
		DBG("%s %d %d %s %d %d ", bss_info->bssid, bss_info->freq,
				bss_info->signal, bss_info->ssid, bss_info->security_type,
				bss_info->encryption_type);

		if (bss_info->ssid[0] == '\0')
			g_free(bss_info);
		else
			bss_info_list = g_slist_append(bss_info_list, bss_info);

	} else
		g_slist_free_full(vsie, g_free);

	return NL_SKIP;
}

static int __netconfig_netlink_scan_reply(struct nl_msg *msg, void *user_data)
{
	/** Called by the kernel when the scan is done or has been aborted. */
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
	struct netconfig_netlink_scan_results *results = user_data;

	if (gnlh->cmd == NL80211_CMD_NEW_SCAN_RESULTS) {
		DBG("Received NL80211_CMD_NEW_SCAN_RESULTS in reply");
		results->done = 1;
		results->aborted = 0;
	} else if (gnlh->cmd == NL80211_CMD_SCAN_ABORTED) {
		DBG("Received NL80211_CMD_SCAN_ABORTED in reply");
		results->done = 1;
		results->aborted = 1;
	}

	return NL_SKIP;
}

static int __netconfig_request_netlink_scan(struct nl_sock *socket,
		int if_index, int id, GVariant *params)
{
	struct netconfig_netlink_scan_results results = { .done = 0, .aborted = 0 };
	struct nl_msg *msg = NULL;
	struct nl_cb *cb = NULL;
	struct nl_msg *ssids = NULL;
	int err = 0;
	int ret = 0;
	unsigned char ies[NETCONFIG_MAX_VSIE_LEN+1] = {0x00, };
	int ies_len = 0;
	GVariantIter *iter;
	GVariant *value;
	gchar *key;
	gboolean ssid_found = FALSE;
	int mcid = __netconfig_get_multicast_id(socket, "nl80211", "scan");

	ret = nl_socket_add_membership(socket, mcid);
	if (ret < 0) {
		DBG("Failed to add membership, error: (%s)", nl_geterror(-ret));
		return ret;
	}

	msg = nlmsg_alloc();
	if (!msg) {
		DBG("Failed to allocate msg");
		nl_socket_drop_membership(socket, mcid);
		return -ENOMEM;
	}
	ssids = nlmsg_alloc();
	if (!ssids) {
		DBG("Failed to allocate ssids");
		nlmsg_free(msg);
		nl_socket_drop_membership(socket, mcid);
		return -ENOMEM;
	}
	cb = nl_cb_alloc(NL_CB_DEFAULT);
	if (!cb) {
		DBG("Failed to allocate callbacks");
		nlmsg_free(msg);
		nlmsg_free(ssids);
		nl_socket_drop_membership(socket, mcid);
		return -ENOMEM;
	}

	/** Set nl message and callback functions. */
	genlmsg_put(msg, 0, 0, id, 0, 0, NL80211_CMD_TRIGGER_SCAN, 0);
	ret = nla_put_u32(msg, NL80211_ATTR_IFINDEX, if_index);
	if (ret < 0) {
		DBG("Failed to add integer attribute to netlink message, error: (%s)", nl_geterror(-ret));
		goto out;
	}

	g_variant_get(params, "a{sv}", &iter);
	while (g_variant_iter_loop(iter, "{sv}", &key, &value)) {
		if (g_strcmp0(key, "SSID") == 0) {
			if (g_variant_is_of_type(value, G_VARIANT_TYPE_STRING)) {
				char *ssid = g_strdup(g_variant_get_string(value, NULL));
				ssid_found = TRUE;
				DBG("ssid [%s]", ssid);

				ret = nla_put(ssids, 1, strlen(ssid), ssid);
				g_free(ssid);
				if (ret < 0) {
					DBG("Failed to add ssid to netlink message, error: (%s)", nl_geterror(-ret));
					g_variant_iter_free(iter);
					goto out;
				}
			}
		} else if (g_strcmp0(key, "VSIE") == 0) {
			if (g_variant_is_of_type(value, G_VARIANT_TYPE_STRING)) {
				char *vsie = g_strdup(g_variant_get_string(value, NULL));
				int vsie_len = strlen(vsie);
				DBG("vsie: %s vsie_len: %d", vsie, vsie_len);

				ies_len = (vsie_len % 2) ? ((vsie_len / 2) + 1) : (vsie_len / 2);
				__netconfig_hex_str_to_bin(vsie, ies, ies_len);
				g_free(vsie);
			}
		}
	}
	g_variant_iter_free(iter);

	if (!ssid_found) {
		ret = nla_put(ssids, 1, 0, "");
		if (ret < 0) {
			DBG("nla_put error: (%s)", nl_geterror(-ret));
			goto out;
		}
	}
	nla_put_nested(msg, NL80211_ATTR_SCAN_SSIDS, ssids);

	if (ies[0] == NETCONFIG_VENDOR_SPECIFIC_ID && ies[1] >= 4) {
		DBG("ies_len: %d ies: %02x %02x %02x %02x %02x %02x %02x", ies_len,
				ies[0], ies[1], ies[2], ies[3], ies[4], ies[5], ies[6]);
		ret = nla_put(msg, NL80211_ATTR_IE, ies_len, ies);
		if (ret < 0) {
			DBG("Failed to add vsie data to netlink message, error: (%s)", nl_geterror(-ret));
			goto out;
		}
	}

	err = 1;
	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, __netconfig_netlink_scan_reply, &results);
	nl_cb_err(cb, NL_CB_CUSTOM, error_handler, &err);
	nl_cb_set(cb, NL_CB_FINISH, NL_CB_CUSTOM, finish_handler, &err);
	nl_cb_set(cb, NL_CB_ACK, NL_CB_CUSTOM, ack_handler, &err);
	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
	ret = nl_socket_set_nonblocking(socket);
	DBG("Set the non-blocking mode [%d]", ret);

	/** Send NL80211_CMD_TRIGGER_SCAN to start the scan. */
	ret = nl_send_auto_complete(socket, msg);

	if (ret < 0) {
		DBG("nl_send_auto_complete() error: (%s)", nl_geterror(-ret));
		goto out;
	}

	DBG("Sent %d bytes to the kernel", ret);
	ssid_found = FALSE;

	while (err > 0) {
		ret = nl_recvmsgs(socket, cb);
		if (ret < 0) {
			DBG("nl_recvmsgs() ret: %d (%s)", ret, nl_geterror(-ret));
			goto out;
		}
	}

	while (!results.done) {
		ret = nl_recvmsgs(socket, cb);
		if (ret < 0 || err < 0) {
			DBG("scan failed");
			ret = -1;
			goto out;
		}
	}

	if (results.aborted) {
		DBG("scan aborted");
		ret = -1;
		goto out;
	}

out:
	/** Release memory */
	nlmsg_free(ssids);
	nlmsg_free(msg);
	nl_cb_put(cb);
	nl_socket_drop_membership(socket, mcid);

	if (ret < 0 || ret == 1)
		return ret;

	DBG("Scan done");
	return 0;
}

static int __netconfig_initialize_nl80211(netconfig_nl_global *global)
{
	int err = 0;

	global->if_index = __netconfig_get_interface_index(WIFI_IFNAME);
	if (global->if_index < 0) {
		DBG("Failed to get interface index");
		return -1;
	}

	global->socket = nl_socket_alloc();
	if (!global->socket) {
		DBG("Failed to allocate netlink socket.");
		return -ENOMEM;
	}

	if (genl_connect(global->socket)) {
		DBG("Failed to connect to generic netlink.");
		err = -ENOLINK;
		goto fail;
	}

	global->id = genl_ctrl_resolve(global->socket, "nl80211");
	if (global->id < 0) {
		DBG("Failed to find the nl80211 driver");
		err = -ENOENT;
		goto fail;
	}

	return 0;

fail:
	nl_socket_free(global->socket);
	return err;
}

static int __netconfig_initialize_nl_msg(netconfig_nl_global *global)
{
	int rv;

	if (global == NULL) {
		DBG("Invalid parameter.");
		return -EINVAL;
	}

	global->msg = nlmsg_alloc();
	if (global->msg == NULL) {
		DBG("Failed to allocate netlink message");
		return -ENOMEM;
	}

	/* Set command into message */
	genlmsg_put(global->msg, 0, 0, global->id, 0, NLM_F_DUMP, NL80211_CMD_GET_SCAN, 0);
	rv = nla_put_u32(global->msg, NL80211_ATTR_IFINDEX, global->if_index);
	if (rv < 0) {
		DBG("Failed to add integer attribute to netlink message.");
		nlmsg_free(global->msg);
		return rv;
	}
	nl_socket_modify_cb(global->socket, NL_CB_VALID, NL_CB_CUSTOM, __netconfig_netlink_scan_cb, NULL);

	return 0;
}

int handle_netlink_scan(Wifi *wifi, GDBusMethodInvocation *context, GVariant *params)
{
	DBG("");
	netconfig_nl_global global = {
		.id = -1,
		.if_index = -1,
		.socket = NULL,
		.msg = NULL,
	};

	/** Initialize netlink socket */
	int ret = __netconfig_initialize_nl80211(&global);
	if (ret < 0) {
		DBG("__netconfig_initialize_nl80211() failed, error %d", ret);
		netconfig_error_dbus_method_return(context, NETCONFIG_ERROR_INTERNAL, "AccessDenied");
		wifi_complete_netlink_scan(wifi, context);
		return TRUE;
	}

	/** Request NL80211_CMD_TRIGGER_SCAN to the kernel. */
	ret = __netconfig_request_netlink_scan(global.socket, global.if_index, global.id, params);
	if (ret < 0) {
		DBG("__netconfig_request_netlink_scan() failed, error %d", ret);
		netconfig_error_dbus_method_return(context, NETCONFIG_ERROR_INTERNAL, "AccessDenied");
		wifi_complete_netlink_scan(wifi, context);
		return TRUE;
	}

	ret = __netconfig_initialize_nl_msg(&global);
	if (ret < 0) {
		DBG("__netconfig_initialize_nl_msg() failed, error %d", ret);
		netconfig_error_dbus_method_return(context, NETCONFIG_ERROR_INTERNAL, "AccessDenied");
		wifi_complete_netlink_scan(wifi, context);
		return TRUE;
	}

	ret = nl_send_auto_complete(global.socket, global.msg);
	DBG("NL80211_CMD_GET_SCAN sent %d bytes to the kernel", ret);

	/** Receive the kernel message. */
	ret = nl_recvmsgs_default(global.socket);
	nlmsg_free(global.msg);
	if (ret < 0) {
		DBG("nl_recvmsgs_default() failed. ret: %d (error: %s)", ret, nl_geterror(-ret));
		netconfig_error_dbus_method_return(context, NETCONFIG_ERROR_INTERNAL, "AccessDenied");
		wifi_complete_netlink_scan(wifi, context);
		return TRUE;
	}

	wifi_complete_netlink_scan(wifi, context);
	__netconfig_notify_netlink_scan_done();

	return TRUE;
}
